import logging

from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/wga/apiac/policy"
requires_modules = ["wga"]
requires_version = "9.0.7"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the list of all existing API Access Control Policies
    """
    return isamAppliance.invoke_get("Retrieving the list of all existing API Access Control Policies",
                                    "{0}".format(uri),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, policy_name, check_mode=False, force=False):
    """
    Retrieve a single API Access Control Policy

    """
    return isamAppliance.invoke_get("Retrieve a single API Access Control Policy",
                                    "{0}/{1}".format(uri, policy_name),
                                    requires_modules=requires_modules, requires_version=requires_version)


def add(isamAppliance, name, groups=None, attributes=None, check_mode=False, force=False):
    """
    Create a new API Access Control Policy
    """
    exist, warnings = _check_exist(isamAppliance, name)

    if force is True or exist is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Create a new API Access Control Policy", uri,
                {
                    "name": name,
                    "groups": groups,
                    "attributes": attributes
                }, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, policy_name, groups=[], attributes=[], check_mode=False, force=False):
    """
    Update an existing API Access Control Policy
    """
    update_required, warnings, json_data = _check(isamAppliance, policy_name, groups, attributes)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update an existing API Access Control Policy",
                "{0}/{1}".format(uri, policy_name),
                json_data,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def set(isamAppliance, policy_name, groups=[], attributes=[], check_mode=False, force=False):
    exist, warnings = _check_exist(isamAppliance, policy_name)
    if exist:
        return update(isamAppliance=isamAppliance, policy_name=policy_name, groups=groups, attributes=attributes,
                      check_mode=check_mode, force=force)
    else:
        return add(isamAppliance=isamAppliance, name=policy_name, groups=groups, attributes=attributes)


def delete(isamAppliance, policy_name, check_mode=False, force=False):
    """
    Delete an access policy
    """
    if force is False:
        exist, warnings = _check_exist(isamAppliance, policy_name)

    if force is True or exist is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete an access policy",
                "{0}/{1}".format(uri, policy_name),
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def delete_all(isamAppliance, check_mode=False, force=False):
    """
    Delete all existing API Access Control Policies
    """
    ret_obj = get_all(isamAppliance)

    if force is True or ret_obj['data'] != []:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete all existing API Access Control Policies",
                "{0}".format(uri),
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete_selection(isamAppliance, policies, command="DELETE", check_mode=False, force=False):
    """
    Delete a selection of API Access Control Policies
    """
    found_any = False
    new_list_policies = []

    ret_obj = get_all(isamAppliance)
    warnings = ret_obj['warnings']

    for policy in policies:
        found = False
        for obj in ret_obj['data']:
            if obj['name'] == policy:
                found_any = True
                found = True
        if found is False:
            warnings.append("Did not find policy {0} to delete".format(policy))
        else:
            new_list_policies.append(policy)

    if force is True or found_any is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Delete a selection of API Access Control Policies",
                "{0}".format(uri),
                {
                    'command': command,
                    'policies': new_list_policies
                },
                requires_modules=requires_modules,
                requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def _check_exist(isamAppliance, policy_name):
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if obj['name'] == policy_name:
            return True, ret_obj['warnings']

    return False, ret_obj['warnings']


def _check(isamAppliance, policy_name, groups, attributes):
    """
    The logic of the get_all() rest api is clunky at best, resulting always in failed checks when using groups or attributes
    - Extended attributes : output always includes quotes, while input does not work if you include quotes (api works, but the result does not)
    - Groups : additional groups ('iv-admin', 'webseal-servers') are added to the output.  The groups are put in ACL entries
    """
    json_data = {}
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if obj['name'] == policy_name:
            obj1 = {
                'name': obj['name'],
            }
            if 'attributes' in obj:
                # parse attributes to match the input (no quotes)
                newAttributes = []
                for l in obj['attributes']:
                    newAttributes.append(l.replace("'", ""))
                obj1["attributes"] = newAttributes
            if 'groups' in obj:
                # groups is not placed in the POP, but in a separate ACL
                #
                defaultGroups = ['iv-admin', 'webseal-servers']
                obj1["groups"] = list(filter(lambda g: g not in defaultGroups, obj["groups"]))
            obj2 = {
                'name': policy_name,
                'groups': groups,
                'attributes': attributes
            }
            sorted_obj1 = tools.json_sort(obj1)
            logger.debug("Sorted existing data: {0}".format(sorted_obj1))
            sorted_obj2 = tools.json_sort(obj2)
            logger.debug("Sorted input data   : {0}".format(sorted_obj2))
            if sorted_obj1 != sorted_obj2:
                logger.info("Changes detected, update needed.")
                return True, ret_obj['warnings'], obj2

    return False, ret_obj['warnings'], json_data


def compare(isamAppliance1, isamAppliance2):
    """
    Compare access policies between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    return tools.json_compare(ret_obj1, ret_obj2)



import logging

logger = logging.getLogger(__name__)


def get(isamAppliance, id, check_mode=False, force=False):
    """
    Retrieve the value of a property in the cluster configuration
    """
    return isamAppliance.invoke_get("Retrieve the value of a property in the cluster configuration",
                                    "/isam/cluster/property/{0}/v2".format(id))


def add(isamAppliance, id, value, check_mode=False, force=False):
    """
    Create a new property in the cluster configuration
    """
    if force is True or _check(isamAppliance, id) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Create a new property in the cluster configuration",
                "/isam/cluster/property/v2",
                {
                    "id": id,
                    "value": value
                })

    return isamAppliance.create_return_object()


def update(isamAppliance, id, value, check_mode=False, force=False):
    """
    Update the value of a property in the cluster configuration
    """
    if force is True or _check(isamAppliance, id, value) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update the value of a property in the cluster configuration",
                "/isam/cluster/property/{0}/v2".format(id),
                {
                    "value": value
                })

    return isamAppliance.create_return_object()


def _check(isamAppliance, id, value=None):
    ret_obj = None
    try:
        ret_obj = get(isamAppliance, id)
    except:
        pass

    if ret_obj is None:
        return False
    else:
        if value is not None and value != ret_obj['data']['value']:
            return False

    return True


def delete(isamAppliance, id, check_mode=False, force=False):
    """
    Delete a property from the cluster configuration
    """
    if force is True or _check(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a property from the cluster configuration",
                "/isam/cluster/property/{0}/v2".format(id))

    return isamAppliance.create_return_object()



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/isam/application_logs"
requires_model = "Appliance"


def get_all(isamAppliance, file_path='', recursive='yes', flat_details=None, check_mode=False, force=False):
    """
    Retrieving the contents of a directory from the application log files area

    flat_details - takes 'yes' or 'no'
    file_path - use the "path" value obtained when run with flat_details = 'yes'
    """
    return isamAppliance.invoke_get("Retrieving the contents of a directory from the application log files area",
                                    "{0}/{1}{2}".format(uri, file_path,
                                                        tools.create_query_string(
                                                            recursive=recursive,
                                                            flat_details=flat_details)), requires_model=requires_model)


def get(isamAppliance, file_path, length=None, start=None, check_mode=False, force=False):
    """
    Retrieving the contents of a file from the application log files area
    """
    return isamAppliance.invoke_get("Retrieving the contents of a file from the application log files area",
                                    "{0}/{1}{2}".format(uri, file_path,
                                                        tools.create_query_string(
                                                            length=length,
                                                            start=start)), requires_model=requires_model)


def clear(isamAppliance, file_id, check_mode=False, force=False):
    """
    Clearing a log file
    """
    ret_obj = {'warnings': ''}

    try:
        ret_obj = get(isamAppliance, file_id, length=1, start=1)
        delete_required = True  # Exception thrown if the file is empty

        # Check for Docker
        warnings = ret_obj['warnings']
        if warnings and 'Docker' in warnings[0]:
            return isamAppliance.create_return_object(warnings=ret_obj['warnings'])
    except:
        delete_required = False

    if force is True or delete_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=ret_obj['warnings'])
        else:
            return isamAppliance.invoke_delete(
                "Clearing a log file",
                "{0}/{1}?action=clear".format(uri, file_id), requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=ret_obj['warnings'])


def delete(isamAppliance, file_id, check_mode=False, force=False):
    """
    Deleting a log file
    """
    delete_required = False
    ret_obj = {'warnings': ''}

    try:
        ret_obj = get(isamAppliance, file_id, length=1, start=1)
        delete_required = True  # Exception thrown if the file is empty

        # Check for Docker
        warnings = ret_obj['warnings']
        if warnings and 'Docker' in warnings[0]:
            return isamAppliance.create_return_object(warnings=ret_obj['warnings'])
    except:
        delete_required = False

    if force is True or delete_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=ret_obj['warnings'])
        else:
            return isamAppliance.invoke_delete(
                "Deleting a log file",
                "{0}/{1}?action=delete".format(uri, file_id), requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=ret_obj['warnings'])


def export_file(isamAppliance, file_path, filename, check_mode=False, force=False):
    """
    Downloading a file from the file application log files area
    """
    import os.path
    ret_obj = {'warnings': ''}

    if force is True or (os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Downloading a file from the file application log files area",
                "{0}/{1}?type=File".format(uri, file_path),
                filename, no_headers=True, requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=ret_obj['warnings'])



import logging
import ibmsecurity.utilities.tools

try:
    basestring
except NameError:
    basestring = (str, bytes)

logger = logging.getLogger(__name__)

# URI for this module
uri = "/isam/runtime"


def get_all(isamAppliance, resource_id, stanza_id, check_mode=False, force=False):
    """
    Retrieving all configuration entries for a stanza - Runtime Environment
    """
    try:
        return isamAppliance.invoke_get("Retrieving all configuration entries for a stanza - Runtime Environment",
                                        "{0}/{1}/configuration/stanza/{2}".format(uri,
                                                                                  resource_id,
                                                                                  stanza_id))
    except:
        # Return empty array - exception thrown if stanza has no entries or does not exist
        ret_obj = isamAppliance.create_return_object()
        ret_obj['data'] = {}
        return ret_obj


def get(isamAppliance, resource_id, stanza_id, entry_id, check_mode=False, force=False):
    """
    Retrieving a specific configuration entry - Runtime Environment
    """
    return isamAppliance.invoke_get("Retrieving a specific configuration entry - Runtime Environment",
                                    "{0}/{1}/configuration/stanza/{2}/entry_name/{3}".format(uri,
                                                                                             resource_id,
                                                                                             stanza_id,
                                                                                             entry_id))


def add(isamAppliance, resource_id, stanza_id, entries, check_mode=False, force=False):
    """
    Adding a configuration entry or entries by stanza - Runtime Environment
    """
    if (isinstance(entries, basestring)):
        import ast
        entries = ast.literal_eval(entries)

    exists = False

    if force is False:
        for entry in entries:
            exists, update_required, value = _check(isamAppliance, resource_id, stanza_id, entry[0], entry[1])
            if exists is True:
                logger.info('Atleast one of the entries being added exists already {0}/{1}/{2}/{3}!'.format(resource_id,
                                                                                                            stanza_id,
                                                                                                            entry[0],
                                                                                                            entry[1]))
                break

    if force is True or exists is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return _add(isamAppliance, resource_id, stanza_id, entries)

    return isamAppliance.create_return_object()


def _add(isamAppliance, resource_id, stanza_id, entries):
    return isamAppliance.invoke_post(
        "Adding a configuration entry or entries by stanza - Runtime Environment",
        "/isam/runtime/{0}/configuration/stanza/{1}/entry_name".format(resource_id, stanza_id),
        {'entries': entries})


def set(isamAppliance, resource_id, stanza_id, entries, check_mode=False, force=False):
    """
    Set a configuration entry or entries by stanza - Runtime Environment

    Note: entries has to be [['key', 'value1'], ['key', 'value2]], cannot provide [['key', ['value1', 'value2']]]
    get() returns the second format - thus lots of logic to handle this discrepancy.

    Smart enough to update only that which is needed.
    """
    set_update = False
    set_entries = []

    if force is False:
        for entry in _collapse_entries(entries):
            process_entry = False
            exists, update_required, cur_value = _check(isamAppliance, resource_id, stanza_id, entry[0], entry[1])
            if exists is False:  # Missing entries
                set_update = True
                process_entry = True
            elif update_required is True:  # Existing entries need to be updated
                set_update = True
                process_entry = True
                # Force delete of existing values, new values will be added
                logger.info(
                    'Deleting entry, will be re-added: {0}/{1}/{2}'.format(resource_id, stanza_id, entry[0]))
                delete_all(isamAppliance, resource_id, stanza_id, entry[0], check_mode, True)
            if process_entry is True:
                if isinstance(entry[1], list):
                    for v in entry[1]:
                        set_entries.append([entry[0], v])
                else:
                    set_entries.append([entry[0], entry[1]])

    if force is True or set_update is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return _add(isamAppliance, resource_id, stanza_id, set_entries)

    return isamAppliance.create_return_object()


def _collapse_entries(entries):
    """
    Convert [['key', 'value1'], ['key', 'value2]] to [['key', ['value1', 'value2']]]
    Expect key values to be consecutive. Will maintain order as provided.
    """
    if entries is None or len(entries) < 1:
        return []
    else:
        cur_key = entries[0][0]
        cur_value = []
        new_entry = []

    for entry in entries:
        if entry[0] == cur_key:
            cur_value.append(entry[1])
        else:
            new_entry.append([cur_key, cur_value])
            cur_key = entry[0]
            cur_value = [entry[1]]

    new_entry.append([cur_key, cur_value])

    return new_entry


def delete(isamAppliance, resource_id, stanza_id, entry_id, value_id='', check_mode=False, force=False):
    """
    Deleting a value from a configuration entry - Runtime Environment
    """
    if force is False:
        exists, update_required, value = _check(isamAppliance, resource_id, stanza_id, entry_id, value_id)

    if force is True or exists is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            # URL being encoded primarily to handle spaces in them
            f_uri = "{0}/{1}/configuration/stanza/{2}/entry_name/{3}/value/{4}".format(uri, resource_id, stanza_id,
                                                                                       entry_id, value_id)
            try:
                # Assume Python3 and import package
                from urllib.parse import quote
            except ImportError:
                # Now try to import Python2 package
                from urllib import quote
            full_uri = quote(f_uri)
            return isamAppliance.invoke_delete(
                "Deleting a value from a configuration entry - Runtime Environment", full_uri)

    return isamAppliance.create_return_object()


def delete_all(isamAppliance, resource_id, stanza_id, entry_id, check_mode=False, force=False):
    """
    Deleting all values from a configuration entry - Runtime Environment
    """
    delete_required = False
    if force is False:
        try:
            ret_obj = get(isamAppliance, resource_id, stanza_id, entry_id)
            if ret_obj['data'] != {}:
                delete_required = True
        except:
            pass

    if force is True or delete_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            f_uri = "{0}/{1}/configuration/stanza/{2}/entry_name/{3}".format(uri, resource_id, stanza_id, entry_id)
            return isamAppliance.invoke_delete(
                "Deleting all values from a configuration entry - Runtime Environment", f_uri)

    return isamAppliance.create_return_object()


def update(isamAppliance, resource_id, stanza_id, entry_id, value_id, check_mode=False, force=False):
    """
    Updating a configuration entry or entries by stanza - Runtime Environment
    """
    if force is False:
        exists, update_required, value = _check(isamAppliance, resource_id, stanza_id, entry_id, value_id)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Updating a configuration entry or entries by stanza - Runtime Environment",
                "{0}/{1}/configuration/stanza/{2}/entry_name/{3}".format(uri,
                                                                         resource_id,
                                                                         stanza_id,
                                                                         entry_id),
                {
                    'value': value_id
                })

    return isamAppliance.create_return_object()


def _check(isamAppliance, resource_id, stanza_id, entry_id, value_id):
    """
    Check if entry/value exists
    """
    try:
        ret_obj = get(isamAppliance, resource_id, stanza_id, entry_id)
    except:
        return False, True, None  # Exception means entry / stanza not found

    exists = True
    update_required = False
    value = ret_obj['data'][entry_id]
    logger.info("Entry found in resource:{0}, stanza:{1}, entryid:{2}, value:{3}".format(resource_id,
                                                                                         stanza_id,
                                                                                         entry_id,
                                                                                         value))
    logger.debug("Existing Value(s): {0}".format(value))
    logger.debug("Value to update  : {0}".format(value_id))

    if isinstance(value_id, list):
        if value != value_id:  # Comparing list with no sorting... sequence of values is of importance
            logger.debug("Value arrays do not match!")
            update_required = True
    else:  # assuming base string provided for value_id
        if len(value) == 1:
            if str(value_id) != str(value[0]):
                logger.debug("Single value do not match!")
                update_required = True
                exists = False  # to satisfy delete call
        else:  # base string will not match a zero length array or multiple values in it
            logger.debug("Current non-single value does not match provided single value!")
            update_required = True
            if value_id not in value:
                logger.debug("Current non-single value does not contain provided single value!")
                exists = False

    return exists, update_required, value


def compare(isamAppliance1, isamAppliance2, resource_id, stanza_id):
    """
    Compare suffixes in two appliances
    """
    ret_obj1 = get_all(isamAppliance1, resource_id, stanza_id)
    ret_obj2 = get_all(isamAppliance2, resource_id, stanza_id)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
import os.path
from ibmsecurity.utilities.tools import files_same, get_random_temp_dir
import shutil

logger = logging.getLogger(__name__)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving all URL Mapping
    """
    return isamAppliance.invoke_get("Retrieving all URL Mapping",
                                    "/wga/dynurl_config")


def get(isamAppliance, id, check_mode=False, force=False):
    """
    Retrieve a URL Mapping
    """
    return isamAppliance.invoke_get("Retrieve a URL Mapping",
                                    "/wga/dynurl_config/{0}".format(id))


def _get_template(isamAppliance):
    """
    Retrieve a URL Mapping Template
    """
    return isamAppliance.invoke_get("Retrieve a URL Mapping Template",
                                    "/wga_templates/dynurl_template.json")


def export_template(isamAppliance, filename, check_mode=False, force=False):
    """
    Exporting the DynURL configuration file template
    """

    if os.path.exists(filename) is True:
        logger.info("File '{0}' already exists.  Skipping export.".format(filename))
        return isamAppliance.create_return_object()

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_get_file(
            "Exporting the DynURL configuration file template",
            "/wga_templates/dynurl_template?export",
            filename)

    return isamAppliance.create_return_object()


def add(isamAppliance, name, dynurl_config_data=None, check_mode=False, force=False):
    """
    Add a URL Mapping
    """
    # Add default template if not specified
    if dynurl_config_data is None:
        ret_obj = _get_template(isamAppliance)
        dynurl_config_data = ret_obj['data']['id']

    if force is True or _check(isamAppliance, name) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Add a URL Mapping",
                "/wga/dynurl_config",
                {
                    "name": name,
                    "dynurl_config_data": dynurl_config_data
                })

    return isamAppliance.create_return_object()


def delete(isamAppliance, id, check_mode=False, force=False):
    """
    Deleting a URL Mapping
    """
    if force is True or _check(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a URL Mapping",
                "/wga/dynurl_config/{0}".format(id))

    return isamAppliance.create_return_object()


def rename(isamAppliance, id, new_name, check_mode=False, force=False):
    """
    Rename a URL Mapping
    """
    if force is True or _check(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Rename a URL Mapping",
                "/wga/dynurl_config/{0}".format(id),
                {
                    'id': id,
                    'new_name': new_name
                })

    return isamAppliance.create_return_object()


def update(isamAppliance, id, dynurl_config_data, check_mode=False, force=False):
    """
    Update a URL Mapping
    """
    if force is True or _check(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a URL Mapping",
                "/wga/dynurl_config/{0}".format(id),
                {
                    'id': id,
                    'dynurl_config_data': dynurl_config_data
                })

    return isamAppliance.create_return_object()


def export_file(isamAppliance, id, filename, check_mode=False, force=False):
    """
    Exporting a URL Mapping
    """
    if force is True or _check(isamAppliance, id) is True:
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Export a URL Mapping",
                "/wga/dynurl_config/{0}?export".format(id),
                filename)

    return isamAppliance.create_return_object()


def import_file(isamAppliance, id, filename, check_mode=False, force=False):
    """
    Importing a URL Mapping
    """
    if force is True or _check_import(isamAppliance, id, filename, check_mode=check_mode):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Importing a URL Mapping",
                "/wga/dynurl_config.js",
                [
                    {
                        'file_formfield': 'dynurl_config_file',
                        'filename': filename,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {})

    return isamAppliance.create_return_object()


def _check(isamAppliance, id):
    """
    Check if URL Mapping already exists
    """
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if obj['id'] == id:
            return True

    return False


def _check_import(isamAppliance, id, filename, check_mode=False):
    """
    Checks if file on the Appliance exists and if so, whether it is different from filename
    """
    tmpdir = get_random_temp_dir()
    tmp_original_file = os.path.join(tmpdir, os.path.basename(id))
    if _check(isamAppliance, id):
        export_file(isamAppliance, id, tmp_original_file, check_mode=False, force=True)
        logger.debug("file already exists on appliance")
        if files_same(tmp_original_file, filename):
            logger.debug("files are the same, so we don't want to do anything")
            shutil.rmtree(tmpdir)
            return False
        else:
            logger.debug("files are different, so we delete existing file in preparation for import")
            delete(isamAppliance, id, check_mode=check_mode, force=True)
            shutil.rmtree(tmpdir)
            return True
    else:
        logger.debug("file does not exist on appliance, so we'll want to import")
        shutil.rmtree(tmpdir)
        return True


def compare(isamAppliance1, isamAppliance2):
    """
    Compare URL Mapping between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['version']
        ret_obj = get(isamAppliance1, obj['id'])
        obj['script'] = ret_obj['data']['contents']
    for obj in ret_obj2['data']:
        del obj['version']
        ret_obj = get(isamAppliance2, obj['id'])
        obj['script'] = ret_obj['data']['contents']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['version'])



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)
requires_model = "Appliance"

def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve existing suffixes.
    """
    return isamAppliance.invoke_get("Retrieving suffixes from embedded ldap",
                                    "/isam/embedded_ldap/suffixes/v1",requires_model=requires_model)


def add(isamAppliance, name, check_mode=False, force=False):
    """
    Add a suffix to embedded ldap
    """
    exists = False
    id = None

    if force is False:
        exists, id, warnings = _check(isamAppliance, name)

    if force is True or exists is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True,warnings=warnings)
        else:
            return isamAppliance.invoke_post(
                "Add suffix to embedded ldap",
                "/isam/embedded_ldap/suffixes/v1",
                {'name': name},requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Add a suffix to embedded ldap
    """
    exists = False
    id = None

    if force is False:
        exists, id, warnings = _check(isamAppliance, name)

    if force is True or exists is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True,warnings=warnings)
        else:
            return isamAppliance.invoke_delete(
                "Delete suffix from embedded ldap",
                "/isam/embedded_ldap/suffixes/{0}/v1".format(id),requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, name):
    """
    Check if suffix exists
    """
    ret_obj = get(isamAppliance)
    check_value, warnings = False, ret_obj['warnings']

    if warnings == []:
        for suffix in ret_obj['data']:
            if suffix['name'] == name:
                logger.info("Suffix found in embedded ldap: " + name)
                check_value = True
                return check_value, suffix['id'], warnings

        logger.info("Suffix *not* found in embedded ldap: " + name)
    return check_value, None, warnings


def compare(isamAppliance1, isamAppliance2):
    """
    Compare suffixes in two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
import copy
from ibmsecurity.utilities import tools
from ibmsecurity.isam.fed.sts import templates
from ibmsecurity.isam.aac import mapping_rules

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/sts/chains"
requires_modules = ['federation']
requires_version = "9.0.1.0"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of STS chains
    """
    return isamAppliance.invoke_get("Retrieve a list of STS chains", uri,
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search STS Chain by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found STS Chain {0} id: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific STS chain
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        warnings = ["STS Chain {0} had no match, skipping retrieval.".format(name)]
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        return _get(isamAppliance, id)


def _get(isamAppliance, id):
    """
    Internal function to get data using "id" - used to avoid extra calls
    """
    return isamAppliance.invoke_get("Retrieve a specific STS chain", "{0}/{1}".format(uri, id),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def set(isamAppliance, name, chainName, requestType, description=None, tokenType=None, xPath=None, signResponses=None,
        signatureKey=None, validateRequests=None, validationKey=None, sendValidationConfirmation=None, issuer=None,
        appliesTo=None, properties=None, new_name=None, ignore_password_for_idempotency=False, check_mode=False,
        force=False):
    """
    Creating or Modifying an STS Chain
    """
    if (search(isamAppliance, name=name))['data'] == {}:
        # Force the add - we already know Chain does not exist
        logger.info("STS Chain {0} had no match, requesting to add new one.".format(name))
        return add(isamAppliance, name=name, chainName=chainName, requestType=requestType, description=description,
                   tokenType=tokenType, xPath=xPath, signResponses=signResponses, signatureKey=signatureKey,
                   validateRequests=validateRequests, validationKey=validationKey,
                   sendValidationConfirmation=sendValidationConfirmation, issuer=issuer, appliesTo=appliesTo,
                   properties=properties, check_mode=check_mode, force=True)
    else:
        # Update request
        logger.info("STS Chain {0} exists, requesting to update.".format(name))
        return update(isamAppliance, name=name, chainName=chainName, requestType=requestType, description=description,
                      tokenType=tokenType, xPath=xPath, signResponses=signResponses, signatureKey=signatureKey,
                      validateRequests=validateRequests, validationKey=validationKey,
                      sendValidationConfirmation=sendValidationConfirmation, issuer=issuer, appliesTo=appliesTo,
                      properties=properties, new_name=new_name,
                      ignore_password_for_idempotency=ignore_password_for_idempotency, check_mode=check_mode,
                      force=force)


def add(isamAppliance, name, chainName, requestType, description=None, tokenType=None, xPath=None, signResponses=None,
        signatureKey=None, validateRequests=None, validationKey=None, sendValidationConfirmation=None, issuer=None,
        appliesTo=None, properties=None, check_mode=False, force=False):
    """
    Create an STS chain
    """
    warnings = []
    if force is False:
        ret_obj = search(isamAppliance, name)

    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            ret_obj = templates.search(isamAppliance, name=chainName)
            if ret_obj['data'] == {}:
                warnings.append("Unable to find a valid STS Chain Template for {0}".format(chainName))
            else:
                chainId = ret_obj['data']
                json_data = {
                    "name": name,
                    "chainId": chainId,
                    "requestType": requestType
                }
                if description is not None:
                    json_data['description'] = description
                if tokenType is not None:
                    json_data['tokenType'] = tokenType
                if xPath is not None:
                    json_data['xPath'] = xPath
                if signResponses is not None:
                    json_data['signResponses'] = signResponses
                if signatureKey is not None:
                    json_data['signatureKey'] = signatureKey
                if validateRequests is not None:
                    json_data['validateRequests'] = validateRequests
                if validationKey is not None:
                    json_data['validationKey'] = validationKey
                if sendValidationConfirmation is not None:
                    json_data['sendValidationConfirmation'] = sendValidationConfirmation
                if issuer is not None:
                    json_data['issuer'] = issuer
                if appliesTo is not None:
                    json_data['appliesTo'] = appliesTo
                if properties is not None:
                    for idx, x in enumerate(properties['self']):
                        if "map.rule.reference.names" in x['name']:
                            ret_obj1 = mapping_rules.search(isamAppliance, x['value'][0])
                            properties['self'].append(
                                {"name": x['prefix'] + ".map.rule.reference.ids", "value": [ret_obj1['data']]})
                            del properties['self'][idx]
                    json_data['properties'] = properties
                return isamAppliance.invoke_post(
                    "Create an STS chain", uri, json_data,
                    requires_modules=requires_modules,
                    requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete a specific STS chain
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    chain_id = ret_obj['data']

    if chain_id == {}:
        logger.info("STS Chain {0} not found, skipping delete.".format(name))
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a specific STS chain",
                "{0}/{1}".format(uri, chain_id),
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def update(isamAppliance, name, chainName, requestType, description=None, tokenType=None, xPath=None,
           signResponses=None, signatureKey=None, validateRequests=None, validationKey=None,
           sendValidationConfirmation=None, issuer=None, appliesTo=None, properties=None, new_name=None,
           ignore_password_for_idempotency=False, check_mode=False, force=False):
    """
    Update a specific STS chain
    """
    warnings = []
    chain_id, update_required, json_data = _check(isamAppliance, name, chainName, requestType, description, tokenType,
                                                  xPath, signResponses, signatureKey, validateRequests, validationKey,
                                                  sendValidationConfirmation, issuer, appliesTo, properties, new_name,
                                                  ignore_password_for_idempotency)
    if chain_id is None:
        warnings.append("Cannot update data for unknown STS Chain (or template): {0}".format(name))
    else:
        if force is True or update_required is True:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True)
            else:
                return isamAppliance.invoke_put(
                    "Update a specific STS chain",
                    "{0}/{1}".format(uri, chain_id), json_data,
                    requires_modules=requires_modules,
                    requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, name, chainName, requestType, description, tokenType, xPath, signResponses, signatureKey,
           validateRequests, validationKey, sendValidationConfirmation, issuer, appliesTo, properties, new_name,
           ignore_password_for_idempotency):
    """
    Check and return True if update needed
    """
    update_required = False
    json_data = {
        "name": name,
        "requestType": requestType
    }
    ret_obj = templates.search(isamAppliance, name=chainName)
    if ret_obj['data'] == {}:
        logger.info("Unable to find a valid STS Chain Template for {0}".format(chainName))
        return None, update_required, json_data
    else:
        json_data['chainId'] = ret_obj['data']
    ret_obj = get(isamAppliance, name)
    if ret_obj['data'] == {}:
        logger.info("STS Chain not found, returning no update required.")
        return None, update_required, json_data
    else:
        if (ignore_password_for_idempotency and ret_obj['data']['properties']):
            if (ret_obj['data']['properties']['self'] == []):
                del ret_obj['data']['properties']['self']
            else:
                for idx, x in enumerate(ret_obj['data']['properties']['self']):
                    if "password" in x['name']:
                        del ret_obj['data']['properties']['self'][idx]
            if (ret_obj['data']['properties']['partner'] == []):
                del ret_obj['data']['properties']['partner']
            else:
                for idx, x in enumerate(ret_obj['data']['properties']['partner']):
                    if "password" in x['name']:
                        del ret_obj['data']['properties']['partner'][idx]
        chain_id = ret_obj['data']['id']
        del ret_obj['data']['id']
        if new_name is not None:
            json_data['name'] = new_name
        if description is not None:
            json_data['description'] = description
        elif 'description' in ret_obj['data']:
            del ret_obj['data']['description']
        if tokenType is not None:
            json_data['tokenType'] = tokenType
        if xPath is not None:
            json_data['xPath'] = xPath
        if signResponses is not None:
            json_data['signResponses'] = signResponses
        if signatureKey is not None:
            json_data['signatureKey'] = signatureKey
        if validateRequests is not None:
            json_data['validateRequests'] = validateRequests
        if validationKey is not None:
            json_data['validationKey'] = validationKey
        if sendValidationConfirmation is not None:
            json_data['sendValidationConfirmation'] = sendValidationConfirmation
        if issuer is not None:
            json_data['issuer'] = issuer
        if appliesTo is not None:
            json_data['appliesTo'] = appliesTo
        if properties is not None:
            for idx, x in enumerate(properties['self']):
                if "map.rule.reference.names" in x['name']:
                    ret_obj1 = mapping_rules.search(isamAppliance, x['value'][0])
                    properties['self'].append(
                        {"name": x['prefix'] + ".map.rule.reference.ids", "value": [ret_obj1['data']]})
                    del properties['self'][idx]
            json_data['properties'] = properties

        if ignore_password_for_idempotency:
            temp = copy.deepcopy(
                json_data)  # deep copy neccessary: otherwise password parameter would be removed from desired config dict 'json_data'
            if ('self' in temp['properties']):
                for idx, x in enumerate(temp['properties']['self']):
                    if "password" in x['name']:
                        del temp['properties']['self'][idx]
            if ('partner' in temp['properties']):
                for idx, x in enumerate(temp['properties']['partner']):
                    if "password" in x['name']:
                        del temp['properties']['partner'][idx]
        else:
            temp = json_data
        sorted_json_data = tools.json_sort(temp)
        logger.debug("Sorted input: {0}".format(sorted_json_data))
        sorted_ret_obj = tools.json_sort(ret_obj['data'])
        logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
        if sorted_ret_obj != sorted_json_data:
            logger.info("Changes detected, update needed.")
            update_required = True

    return chain_id, update_required, json_data


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Module Chains between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
    for obj in ret_obj2['data']:
        del obj['id']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id'])



import logging
import os.path

from ibmsecurity.utilities.tools import json_sort

logger = logging.getLogger(__name__)
uri = "/wga/reverseproxy"
requires_modules = "wga"
requires_version = None
requires_model = "Appliance"

def get(isamAppliance, instance_id, check_mode=False, force=False):
    """
    Retrieving all transaction logging components and their details

    """
    return isamAppliance.invoke_get("Retrieving all transaction logging components and their details",
                                    "{0}/{1}/transaction_logging".format(uri, instance_id),
                                    requires_modules=requires_modules, requires_version=requires_version,requires_model=requires_model)


def get_files(isamAppliance, instance_id, component_id, check_mode=False, force=False):
    """
    Retrieving all transaction log files for a component

    """
    return isamAppliance.invoke_get("Retrieving all transaction log files for a component",
                                    "{0}/{1}/transaction_logging/{2}/translog_files".format(uri, instance_id,
                                                                                            component_id),
                                    requires_modules=requires_modules, requires_version=requires_version,requires_model=requires_model)


def export_file(isamAppliance, instance_id, component_id, file_id, filepath, check_mode=False, force=False):
    """
    Exporting the transaction logging data file or rollover transaction logging data file for a component
    """

    if os.path.exists(filepath) is True:
        logger.info("File '{0}' already exists.  Skipping export.".format(filepath))
        warnings = ["File '{0}' already exists.  Skipping export.".format(filepath)]
        return isamAppliance.create_return_object(warnings=warnings)

    warnings = _check(isamAppliance,instance_id,component_id)

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True,warnings=warnings)
    else:
        return isamAppliance.invoke_get_file(
            "Exporting the transaction logging data file or rollover transaction logging data file for a component",
            "{0}/{1}/transaction_logging/{2}/translog_files/{3}?export".format(uri, instance_id, component_id, file_id),
            filepath
        ,requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, instance_id, component_id, status, rollover_size, max_rollover_files, compress,
           check_mode=False, force=False):
    """
    Modifying the status and rollover size for a component

    """

    check_value,warnings = _check_update(isamAppliance, instance_id=instance_id, component_id=component_id, status=status,
                                         rollover_size=rollover_size, max_rollover_files=max_rollover_files,
                                         compress=compress)

    if force is True or check_value is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True,warnings=warnings)
        else:

            return isamAppliance.invoke_put(
                "Modifying the status and rollover size for a component",
                "{0}/{1}/transaction_logging/{2}".format(uri, instance_id, component_id),
                {
                    'id': component_id,
                    'status': status,
                    'rollover_size': rollover_size,
                    'max_rollover_files': max_rollover_files,
                    'compress': compress
                },
                requires_modules=requires_modules, requires_version=requires_version,requires_model=requires_model
            )

    return isamAppliance.create_return_object(warnings=warnings)


def rollover(isamAppliance, instance_id, component_id, check_mode=False, force=False):
    """
    Rolling over the transaction logging data file for a component

    """
    check_value, warnings = _check_enabled(isamAppliance, instance_id, component_id)

    if force is True or check_value is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True,warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Rolling over the transaction logging data file for a component",
                "{0}/{1}/transaction_logging/{2}".format(uri, instance_id, component_id),
                {
                    'rollover': "yes"
                },
                requires_modules=requires_modules, requires_version=requires_version,requires_model=requires_model
            )

    return isamAppliance.create_return_object(warnings=warnings)


def delete_unused(isamAppliance, instance_id, component_id, check_mode=False, force=False):
    """
    Deleting the unused transaction logging data file and rollover files for a component
    """

    ret_obj = get_files(isamAppliance, instance_id=instance_id, component_id=component_id)

    if force is True or ret_obj['data'] != []:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting the unused transaction logging data file and rollover files for a component",
                "{0}/{1}/transaction_logging/{2}/translog_files".format(uri, instance_id, component_id),requires_model=requires_model)

    return isamAppliance.create_return_object()


def delete(isamAppliance, instance_id, component_id, file_id, check_mode=False, force=False):
    """
    Deleting the transaction logging data file or rollover file for a component
    """
    check_value,warnings = _check_file(isamAppliance, instance_id, component_id, file_id)

    if force is True or check_value is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True,warnings=warnings)
        else:
            return isamAppliance.invoke_delete(
                "Deleting the transaction logging data file or rollover file for a component",
                "{0}/{1}/transaction_logging/{2}/translog_files/{3}".format(uri, instance_id, component_id, file_id),requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def delete_multiple_files(isamAppliance, instance_id, component_id, files, check_mode=False, force=False):
    """
    Deleting multiple transaction logging data file and rollover files for a component
    """
    delete_required = False
    files_to_delete = []

    check_value, warnings = _check(isamAppliance,instance_id,component_id)

    if warnings == []:
        ret_obj = get_files(isamAppliance, instance_id, component_id)

        for obj1 in files:
            for obj2 in ret_obj['data']:
                if obj1['name'] == obj2['id']:
                    files_to_delete.append(obj1)
                    delete_required = True

        if len(files_to_delete) == 1:
            return delete(isamAppliance, instance_id, component_id, files_to_delete[0]['name'])

        if force is True or delete_required is True:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True)
            else:
                return isamAppliance.invoke_put(
                    "Deleting multiple transaction logging data file and rollover files for a component",
                    "{0}/{1}/transaction_logging/{2}/translog_files?action=delete".format(uri, instance_id, component_id),
                    {
                        'files': files
                    },requires_model=requires_model
                    )

    return isamAppliance.create_return_object(warnings=warnings)


def _check_update(isamAppliance, instance_id, component_id, status, rollover_size, max_rollover_files, compress):
    """
    Check to see if the file_id exists or not
    """
    ret_obj = get(isamAppliance, instance_id)
    warnings = ret_obj['warnings']
    new_obj = {
        'id': component_id,
        'status': status,
        'rollover_size': rollover_size,
        'max_rollover_files': max_rollover_files,
        'compress': compress
    }

    sorted_new_obj = json_sort(new_obj)

    for obj in ret_obj['data']:
        if obj['id'] == component_id:
            if obj['status'] == "Off" and status == "Off":
                return True,warnings
            else:
                del obj['file_size']
                sorted_obj = json_sort(obj)
                if sorted_obj == sorted_new_obj:
                    return True,warnings

    return False,warnings


def _check_file(isamAppliance, instance_id, component_id, file_id):
    """
    Check to see if the file_id exists or not
    """
    ret_obj = get_files(isamAppliance, instance_id, component_id)
    warnings = ret_obj['warnings']

    for obj in ret_obj['data']:
        if obj['id'] == file_id:
            return True,warnings

    return False,warnings


def _check(isamAppliance, instance_id, component_id):
    """
    Check to see if the component_id exists or not
    """
    ret_obj = get(isamAppliance, instance_id)
    warnings = ret_obj['warnings']
    if warnings == []:
        for obj in ret_obj['data']:
            if obj['id'] == component_id:
                return True,warnings

    return False,warnings


def _check_enabled(isamAppliance, instance_id, component_id):
    """
    Check to see if the component is enabled or not
    """
    ret_obj = get(isamAppliance, instance_id)
    warnings = ret_obj['warnings']

    for obj in ret_obj['data']:
        if obj['id'] == component_id and obj['status'] == "On":
            return True,warnings

    return False,warnings



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/policyattachments"


def get_all(isamAppliance, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of configured resources
    """
    return isamAppliance.invoke_get("Retrieve a list of configured resources",
                                    "{0}{1}".format(uri, tools.create_query_string(filter=filter, sortBy=sortBy)))


def get(isamAppliance, server, resourceUri, check_mode=False, force=False):
    """
    Retrieve a specific configured resource
    """
    ret_obj = search(isamAppliance, server=server, resourceUri=resourceUri, check_mode=check_mode, force=force)
    resource_id = ret_obj['data']

    if resource_id == {}:
        logger.info("Resource {0}/{1} had no match, skipping retrieval.".format(server, resourceUri))
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get("Retrieve a specific configured resource",
                                        "{0}/{1}".format(uri, resource_id))


def get_attachments(isamAppliance, server, resourceUri, check_mode=False, force=False):
    """
    Retrieve a list of attachments for a resource
    """
    ret_obj = search(isamAppliance, server=server, resourceUri=resourceUri, check_mode=check_mode, force=force)
    resource_id = ret_obj['data']

    if resource_id == {}:
        logger.info("Resource {0}/{1} had no match, skipping retrieval.".format(server, resourceUri))
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get("Retrieve a list of attachments for a resource",
                                        "{0}/{1}/policies".format(uri, resource_id))


def search(isamAppliance, server, resourceUri, force=False, check_mode=False):
    """
    Search server/resource uri id by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['resourceUri'] == resourceUri and obj['server'] == server:
            logger.info("Found server/resourceUri {0}/{1} id: {2}".format(server, resourceUri, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def authenticate(isamAppliance, username, password, domain='Default', check_mode=False, force=False):
    """
    Authenticate with Security Access Manager
    """
    ret_obj = isamAppliance.invoke_post(
        "Authenticate with Security Access Manager",
        "{0}/pdadmin".format(uri), {
            "command": "setCredential",
            "username": username,
            "password": password,
            "domain": domain
        })

    ret_obj['changed'] = False  # Authentication call changes nothing on the appliance (HTTP POST defaults to change)

    return ret_obj


def get_resources(isamAppliance, object='/WebSEAL', check_mode=False, force=False):
    """
    Retrieve a list of resources in a protected object space
    """
    ret_obj = isamAppliance.invoke_post(
        "Retrieve a list of resources in a protected object space",
        "{0}/pdadmin".format(uri), {
            "command": "object list",
            "object": object
        })

    ret_obj['changed'] = False  # Resource get call changes nothing on the appliance (HTTP POST defaults to change)

    return ret_obj


def config(isamAppliance, server, resourceUri, policies=[], policyType=None, policyCombiningAlgorithm=None, cache=None,
           check_mode=False, force=False):
    """
    Configure a resource
    Note: Please input policies with policy names (it will be converted to id's), like so:
     [{'name': '<policy name>', 'type': 'policy'}, {'name': '<policyset name>', 'type': 'policyset'},
      {'name': '<definition name>', 'type': 'definition'}]
    """
    warnings = []
    if force is False:
        ret_obj = search(isamAppliance, server, resourceUri)

    if force is True or ret_obj['data'] == {}:
        json_data = {
            "server": server,
            "resourceUri": resourceUri
        }
        if policyType is not None:
            if tools.version_compare(isamAppliance.facts["version"], "9.0.6.0") < 0:
                warnings.append(
                    "Appliance at version: {0}, policyType: {1} is not supported. Needs 9.0.6.0 or higher. Ignoring policyType for this call.".format(
                        isamAppliance.facts["version"], cache))
            else:
                json_data['type'] = policyType

        json_data['policies'] = _convert_policy_name_to_id(isamAppliance, policies)
        if policyCombiningAlgorithm is not None:
            json_data['policyCombiningAlgorithm'] = policyCombiningAlgorithm
        if cache is not None:
            if tools.version_compare(isamAppliance.facts["version"], "9.0.3.0") < 0:
                warnings.append(
                    "Appliance at version: {0}, cache: {1} is not supported. Needs 9.0.3.0 or higher. Ignoring cache for this call.".format(
                        isamAppliance.facts["version"], cache))
            else:
                json_data["cache"] = cache
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post(
                "Configure a resource", uri, json_data, warnings=warnings)

    return isamAppliance.create_return_object()


def update(isamAppliance, server, resourceUri, policyCombiningAlgorithm, cache=None, check_mode=False, force=False):
    """
    Update the policy attachment combining algorithm
    """
    warnings = []
    ret_obj = get(isamAppliance, server, resourceUri)

    if force is True or (
            ret_obj['data'] != {} and (
            ret_obj['data']['policyCombiningAlgorithm'] != policyCombiningAlgorithm or
            ret_obj['data']['cache'] != cache)):
        json_data = {
            "policyCombiningAlgorithm": policyCombiningAlgorithm
        }
        if cache is not None:
            if tools.version_compare(isamAppliance.facts["version"], "9.0.3.0") < 0:
                warnings.append(
                    "Appliance at version: {0}, cache: {1} is not supported. Needs 9.0.3.0 or higher. Ignoring cache for this call.".format(
                        isamAppliance.facts["version"], cache))
            else:
                json_data["cache"] = cache
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Update the policy attachment combining algorithm and cache",
                "{0}/{1}/properties".format(uri, ret_obj['data']['id']), json_data, warnings=warnings)

    return isamAppliance.create_return_object()


def update_attachments(isamAppliance, server, resourceUri, attachments, action, check_mode=False, force=False):
    """
    Update the attachments for a resource

    Provide attachemnts like so:
     [{'name': '<policy name>', 'type': 'policy'}, {'name': '<policyset name>', 'type': 'policyset'},
      {'name': '<definition name>', 'type': 'definition'}]
    """
    ret_obj = get(isamAppliance, server, resourceUri)
    cur_policies = ret_obj['data']['policies']

    if force is True or _check(cur_policies, attachments, action) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            new_policies = _convert_policy_name_to_id(isamAppliance, attachments)
            return isamAppliance.invoke_put(
                "Update the attachments for a resource",
                "{0}/{1}/policies{2}".format(uri, ret_obj['data']['id'], tools.create_query_string(action=action)),
                new_policies)
    return isamAppliance.create_return_object()


def _check(policies, attachments, action):
    # Check and see if there is even one match
    full_match = True
    partial_match = False
    for new_pol in attachments:
        match = False
        for cur_pol in policies:
            if new_pol['name'] == cur_pol['name'] and new_pol['type'] == cur_pol['type']:
                logger.info("At least one policy already exists {0}/{1}".format(new_pol['type'], new_pol['name']))
                match = True
                break
        if match is False:
            full_match = False
        elif partial_match is False:
            partial_match = True

    # Add will be rejected if there is even one match
    if action == 'add':
        logger.info("Check if there is at least one match returned: {0}".format(partial_match))
        return not partial_match
    # Delete will be rejected if there is a partial match (has to be full match)
    elif action == 'remove':
        logger.info("Check if there is a full match of provided policies: {0}".format(full_match))
        return full_match
    # Set requires that there be a full match and the number of elements match
    elif action == 'set':
        if len(policies) == len(attachments) and full_match is True:
            return False
        else:
            return True
    #Force delete will be rejected if there is no match
    elif action == 'force_remove':
        logger.info("Check if there is at least one match returned: {0}".format(partial_match))
        return partial_match
    else:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Unknown action provided: {0}".format(action))


def publish(isamAppliance, server, resourceUri, check_mode=False, force=False):
    """
    Publish the policy attachments for a resource
    """
    ret_obj = get(isamAppliance, server, resourceUri)

    if force is True or (ret_obj['data'] != {} and
                            (ret_obj['data']['deployrequired'] is True or ret_obj['data']['deployed'] is False) and
                             ret_obj['data']['policies'] != []
                        ):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Publish the policy attachments for a resource",
                "{0}/deployment/{1}".format(uri, ret_obj['data']['id']), {})

    return isamAppliance.create_return_object()


def publish_list(isamAppliance, attachments, check_mode=False, force=False):
    """
    Publish a list of policy attachments

    Note: provide attachments like so:
    [{'server': '<server1>', 'resourceUri': '<resourceuri1>'}, {'server': '<server2>', 'resourceUri': '<resourceuri2>'}]
    """
    id_list = []
    for attach in attachments:
        ret_obj = get(isamAppliance, attach['server'], attach['resourceUri'])
        if force is True or ret_obj['data']['deployrequired'] is True:
            id_list.append(ret_obj['data']['id'])
    logger.debug('Attachments: {0}'.format(id_list))

    if len(id_list) > 0:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Publish a list of policy attachments",
                "{0}/deployment".format(uri), {
                    'policyAttachmentIds': ','.join(id_list)
                })

    return isamAppliance.create_return_object()


def _convert_policy_name_to_id(isamAppliance, policies):
    """
    Converts this:
    [{'name': '<policy name>', 'type': 'policy'}, {'name': '<policyset name>', 'type': 'policyset'},
     {'name': '<definition name>}, 'type': 'definition'}]
    to:
    [{'id': '<policy id>', 'type': 'policy'}, {'id': '<policyset id>', 'type': 'policyset'},
     {'id': '<definition id>, 'type': 'definition'}]
    """
    pol_ids = []
    import ibmsecurity.isam.aac.access_control.policies
    import ibmsecurity.isam.aac.access_control.policy_sets
    import ibmsecurity.isam.aac.api_protection.definitions
    for pol in policies:
        name = pol['name']
        type = pol['type']
        if type == 'policy':
            ret_obj = ibmsecurity.isam.aac.access_control.policies.search(isamAppliance, name)
            pol_id = ret_obj['data']
            if pol_id != {}:
                logger.debug("Converting policy {0} to ID: {1}".format(name, pol_id))
            else:
                logger.warning("Unable to find policy {0}, skipping.".format(name))
        elif type == 'policyset':
            ret_obj = ibmsecurity.isam.aac.access_control.policy_sets.search(isamAppliance, name)
            pol_id = ret_obj['data']
            if pol_id != {}:
                logger.debug("Converting policy set {0} to ID: {1}".format(name, pol_id))
            else:
                logger.warning("Unable to find policy set {0}, skipping.".format(name))
        elif type == 'definition':
            ret_obj = ibmsecurity.isam.aac.api_protection.definitions.search(isamAppliance, name)
            pol_id = ret_obj['data']
            if pol_id != {}:
                logger.debug("Converting api definition {0} to ID: {1}".format(name, pol_id))
            else:
                logger.warning("Unable to find api definition {0}, skipping.".format(name))
        else:
            from ibmsecurity.appliance.ibmappliance import IBMError
            raise IBMError("999", "Policy specified with unknown type: {0}/{1}".format(type, name))

        pol_ids.append({'id': pol_id, 'type': type})

    return pol_ids


def _convert_policy_id_to_name(isamAppliance, policies):
    """
    Converts this:
    [{'id': '<policy id>', 'type': 'policy'}, {'id': '<policyset id>', 'type': 'policyset'},
     {'id': '<definition id>, 'type': 'definition'}]
    to:
    [{'name': '<policy name>', 'type': 'policy'}, {'name': '<policyset name>', 'type': 'policyset'},
     {'name': '<definition name>}, 'type': 'definition'}]
    """
    pol_ids = []
    import ibmsecurity.isam.aac.access_control.policies
    import ibmsecurity.isam.aac.access_control.policy_sets
    import ibmsecurity.isam.aac.api_protection.definitions
    for pol in policies:
        pol_id = pol['id']
        type = pol['type']
        if type == 'policy':
            ret_obj = ibmsecurity.isam.aac.access_control.policies._get(isamAppliance, pol_id)
            pol_name = ret_obj['data']['name']
            if pol_name != {}:
                logger.debug("Converting policy {0} to Name: {1}".format(pol_id, pol_name))
            else:
                logger.warning("Unable to find policy {0}, skipping.".format(pol_id))
        elif type == 'policyset':
            ret_obj = ibmsecurity.isam.aac.access_control.policy_sets._get(isamAppliance, pol_id)
            pol_name = ret_obj['data']['name']
            if pol_name != {}:
                logger.debug("Converting policy set {0} to Name: {1}".format(pol_id, pol_name))
            else:
                logger.warning("Unable to find policy set {0}, skipping.".format(pol_id))
        elif type == 'definition':
            ret_obj = ibmsecurity.isam.aac.api_protection.definitions._get(isamAppliance, pol_id)
            pol_name = ret_obj['data']['name']
            if pol_name != {}:
                logger.debug("Converting api definition {0} to Name: {1}".format(pol_id, pol_name))
            else:
                logger.warning("Unable to find api definition {0}, skipping.".format(pol_id))
        else:
            from ibmsecurity.appliance.ibmappliance import IBMError
            raise IBMError("999", "Policy specified with unknown type: {0}/{1}".format(type, pol_id))

        pol_ids.append({'name': pol_name, 'type': type})

    return pol_ids


def delete(isamAppliance, server, resourceUri, check_mode=False, force=False):
    """
    Delete a configured resource
    """
    ret_obj = search(isamAppliance, server, resourceUri, check_mode=check_mode, force=force)
    resourceID = ret_obj['data']

    if resourceID == {}:
        logger.info("ResourceURI {0}/{1} not found, skipping delete.".format(server, resourceUri))
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a configured resource",
                "{0}/{1}".format(uri, resourceID))

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Policy Sets between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
        del obj['userlastmodified']
        del obj['lastmodified']
        del obj['datecreated']
        del obj['lastdeployed']
        for pol in obj['policies']:
            del pol['id']
            del pol['lastmodified']
    for obj in ret_obj2['data']:
        del obj['id']
        del obj['userlastmodified']
        del obj['lastmodified']
        del obj['datecreated']
        del obj['lastdeployed']
        for pol in obj['policies']:
            del pol['id']
            del pol['lastmodified']

    return tools.json_compare(ret_obj1, ret_obj2,
                              deleted_keys=['id', 'userlastmodified', 'lastmodified', 'datecreated', 'lastdeployed'])



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)
uri = "/isam/authzserver"
requires_model = "Appliance"


def get(isamAppliance, id, check_mode=False, force=False):
    """
    Retrieving a list of stanzas - Authorization Server
    """
    return isamAppliance.invoke_get(description="Retrieving a list of stanzas - Authorization Server",
                                    uri="{0}/{1}/configuration/stanza/v1".format(uri, id),
                                    requires_model=requires_model)


def add(isamAppliance, id, stanza_id, check_mode=False, force=False):
    """
    Add Stanza name - Authorization Server
    """
    stanza_exists, warnings = _check(isamAppliance, id, stanza_id)

    if not warnings:
        if force is True or stanza_exists is False:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True, warnings=warnings)
            else:
                return isamAppliance.invoke_post(
                    description="Add stanza name - Authorization Server",
                    uri="{0}/{1}/configuration/stanza/{2}/v1".format(uri, id, stanza_id),
                    data={}, requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, id, stanza_id, check_mode=False, force=False):
    """
    Delete Authorization Server configuration stanza
    """
    stanza_exists, warnings = _check(isamAppliance, id, stanza_id)

    if not warnings:
        if force is True or stanza_exists is True:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True, warnings=warnings)
            else:
                return isamAppliance.invoke_delete(
                    description="Delete Authorization Server configuration stanza",
                    uri="{0}/{1}/configuration/stanza/{2}/v1".format(uri, id, stanza_id),
                    requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, id, stanza_id):
    """
    Check if stanza exists
    """
    ret_obj = get(isamAppliance, id)
    stanza_exists, warnings = False, ret_obj['warnings']

    if not warnings:
        for stanza in ret_obj['data']:
            if stanza == stanza_id:
                logger.info("Stanza found in resource: " + id)
                stanza_exists = True
        logger.info("Stanza *not* found in resource: " + id)
    return stanza_exists, warnings


def compare(isamAppliance1, isamAppliance2, id, id2=None):
    """
    Compare stanzas within authorization server configuration between two appliances
    """
    if id2 is None or id2 == '':
        id2 = id

    import ibmsecurity.isam.web.authorization_server.configuration.entry

    # The following array contains entries that will be ignored across all stanzas
    ignore_entries = ['azn-app-host', 'azn-server-name', 'pd-user-pwd', 'bind-pwd', 'network-interface', 'server-name',
                      'listen-interface']

    # Retrieve all stanzas and corresponding entries for comparison
    ret_obj1 = get(isamAppliance1, id)
    warnings = ret_obj1['warnings']
    if warnings and 'Docker' in warnings[0]:
        return isamAppliance.create_return_object(warnings=ret_obj1['warnings'])

    entries = {}
    for stanza in ret_obj1['data']:
        entries[stanza] = {}
        stanza_entries = ibmsecurity.isam.web.authorization_server.configuration.entry.get_all(isamAppliance1, id,
                                                                                               stanza)
        for k, v in stanza_entries['data'].items():
            if k not in ignore_entries:
                entries[stanza][str(k)] = v
    ret_obj1['data'] = entries

    ret_obj2 = get(isamAppliance2, id2)
    entries = {}
    for stanza in ret_obj2['data']:
        entries[stanza] = {}
        stanza_entries = ibmsecurity.isam.web.authorization_server.configuration.entry.get_all(isamAppliance2, id,
                                                                                               stanza)
        for k, v in stanza_entries['data'].items():
            if k not in ignore_entries:
                entries[stanza][str(k)] = v
    ret_obj2['data'] = entries

    return ibmsecurity.utilities.tools.json_compare(ret_obj1=ret_obj1, ret_obj2=ret_obj2, deleted_keys=ignore_entries)



import logging
import ibmsecurity.utilities.tools
from ibmsecurity.utilities.tools import jsonSortedListEncoder
import json

try:
    basestring
except NameError:
    basestring = (str, bytes)

logger = logging.getLogger(__name__)

# URI for this module
uri = "/wga/reverseproxy"


def get_all(isamAppliance, reverseproxy_id, stanza_id, check_mode=False, force=False):
    """
    Retrieving all configuration entries for a stanza - Reverse Proxy
    """
    try:
        ret_obj = isamAppliance.invoke_get("Retrieving all configuration entries for a stanza - Reverse Proxy",
                                           "{0}/{1}/configuration/stanza/{2}".format(uri,
                                                                                     reverseproxy_id,
                                                                                     stanza_id))
        logger.debug(f"Get All {stanza_id}:\n {ret_obj}")
    except:
        # Return empty array - exception thrown if stanza has no entries or does not exist
        ret_obj = isamAppliance.create_return_object()
        ret_obj['data'] = {}

    return ret_obj


def get(isamAppliance, reverseproxy_id, stanza_id, entry_id, check_mode=False, force=False):
    """
    Retrieving a specific configuration entry - Reverse Proxy
   """
    # URL being encoded primarily to handle request-log-format that has "%" values in them
    f_uri = "{0}/{1}/configuration/stanza/{2}/entry_name/{3}".format(uri, reverseproxy_id,
                                                                     stanza_id, entry_id)
    # Replace % with %25 if it is not encoded already
    import re
    ruri = re.sub("%(?![0-9a-fA-F]{2})", "%25", f_uri)
    # URL encode
    try:
        # Assume Python3 and import package
        from urllib.parse import quote
    except ImportError:
        # Now try to import Python2 package
        from urllib import quote

    full_uri = quote(ruri)
    return isamAppliance.invoke_get("Retrieving a specific configuration entry - Reverse Proxy",
                                    full_uri)


def add(isamAppliance, reverseproxy_id, stanza_id, entries, check_mode=False, force=False):
    """
   Adding a configuration entry or entries by stanza - Reverse Proxy
   """
    if isinstance(entries, basestring):
        import ast
        entries = ast.literal_eval(entries)

    add_required = False

    if force is False:
        add_entries = []
        for entry in entries:
            exists, update_required, value = _check(isamAppliance, reverseproxy_id, stanza_id, entry[0], entry[1])
            if exists is True:
                logger.debug(
                    'Entries exists {0}/{1}/{2}/{3}! Will be ignored.'.format(reverseproxy_id, stanza_id, entry[0],
                                                                              entry[1]))
            else:
                add_entries.append(entry)
                add_required = True
        entries = add_entries

    if force is True or add_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return _add(isamAppliance, reverseproxy_id, stanza_id, entries)

    return isamAppliance.create_return_object()


def _add(isamAppliance, reverseproxy_id, stanza_id, entries):
    return isamAppliance.invoke_post(
        "Adding a configuration entry or entries by stanza - Reverse Proxy",
        "{0}/{1}/configuration/stanza/{2}/entry_name".format(uri, reverseproxy_id, stanza_id),
        {"entries": entries})


def set(isamAppliance, reverseproxy_id, stanza_id, entries, check_mode=False, force=False):
    """
    Set a configuration entry or entries by stanza - Reverse Proxy

    Note: entries has to be [['key', 'value1'], ['key', 'value2]], cannot provide [['key', ['value1', 'value2']]]
    get() returns the second format - thus lots of logic to handle this discrepancy.

    This version will update all entries in the stanza if a single one changes.  But it is a lot faster because it's
    idempotent.

   """
    if isinstance(entries, basestring):
        import ast
        entries = ast.literal_eval(entries)
    # get all entries for the stanza
    currentEntries = get_all(isamAppliance, reverseproxy_id, stanza_id)
    currentEntries = currentEntries['data']
    # make currentEntries into a similar list of dicts
    newEntries = _collapse_entries_obj(entries)
    # Filter the current entries to only include the requested entry (keys)
    fCurrentEntries = {k: v for k, v in currentEntries.items() if k in newEntries.keys()}
    # compare using json_sort
    newEntriesJSON = json.dumps(newEntries, skipkeys=True, sort_keys=True, cls=jsonSortedListEncoder)
    logger.debug(f"\nSorted Desired  Stanza {stanza_id}:\n\n {newEntriesJSON}\n")
    currentEntriesJSON = json.dumps(fCurrentEntries, skipkeys=True, sort_keys=True, cls=jsonSortedListEncoder)
    logger.debug(f"\nSorted Existing Stanza {stanza_id}:\n\n {currentEntriesJSON}\n")

    if force or (newEntriesJSON != currentEntriesJSON):
        for entry in entries:
            logger.info(f"Deleting entry, will be re-added: {reverseproxy_id}/{stanza_id}/{entry[0]}")
            delete_all(isamAppliance, reverseproxy_id, stanza_id, entry[0], check_mode, True)
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return _add(isamAppliance, reverseproxy_id, stanza_id, entries)

    return isamAppliance.create_return_object()


def _collapse_entries_obj(entries):
    """
   Convert [['key', 'value1'], ['key', 'value2]] to {'key': ['value1', 'value2'], ...}
   the value is an array if multiple values, but a string if it's a single value

   Also converts all values to str for easy compare
   """
    if entries is None or len(entries) < 1:
        return []
    else:
        prev_key = ""
        new_entry = {}

    for entry in entries:
        if entry[0] == prev_key:
            if isinstance(new_entry.get(entry[0], ""), list):
                existing_value = new_entry.get(entry[0], [])
            else:
                existing_value = [new_entry.get(entry[0], "")]
            if isinstance(entry[1], list):
                new_entry[entry[0]] = existing_value + str(entry[1])
            else:
                new_entry[entry[0]] = existing_value + [str(entry[1])]
        else:
            new_entry[entry[0]] = str(entry[1])
        prev_key = entry[0]
    return new_entry


def delete(isamAppliance, reverseproxy_id, stanza_id, entry_id, value_id='', check_mode=False, force=False):
    """
   Deleting a value from a configuration entry - Reverse Proxy
   """

    if value_id == '' or value_id is None:
        return delete_all(isamAppliance=isamAppliance, reverseproxy_id=reverseproxy_id, stanza_id=stanza_id,
                          entry_id=entry_id, check_mode=check_mode, force=force)

    if force is False:
        exists, update_required, value = _check(isamAppliance, reverseproxy_id, stanza_id, entry_id, value_id)

    if force is True or exists is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            # URL being encoded primarily to handle request-log-format that has "%" values in them
            f_uri = "{0}/{1}/configuration/stanza/{2}/entry_name/{3}/value/{4}".format(uri, reverseproxy_id, stanza_id,
                                                                                       entry_id, value_id)
            # Replace % with %25 if it is not encoded already
            import re
            ruri = re.sub("%(?![0-9a-fA-F]{2})", "%25", f_uri)
            # URL encode
            try:
                # Assume Python3 and import package
                from urllib.parse import quote
            except ImportError:
                # Now try to import Python2 package
                from urllib import quote

            full_uri = quote(ruri)
            # Workaround for value_id encoding in 9.0.7.1
            if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], '9.0.7.1') >= 0:
                uri_parts = full_uri.split('/value/')
                uri_parts[1] = uri_parts[1].replace('/', '%2F')
                full_uri = '/value/'.join(uri_parts)
            return isamAppliance.invoke_delete(
                "Deleting a value from a configuration entry - Reverse Proxy", full_uri)

    return isamAppliance.create_return_object()


def delete_all(isamAppliance, reverseproxy_id, stanza_id, entry_id, check_mode=False, force=False):
    """
   Deleting all values from a configuration entry - Reverse Proxy
   """
    delete_required = False
    if force is False:
        try:
            ret_obj = get(isamAppliance, reverseproxy_id, stanza_id, entry_id)
            if ret_obj['data'] != {}:
                delete_required = True
        except:
            pass

    if force is True or delete_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            # URL being encoded primarily to handle request-log-format that has "%" values in them
            f_uri = "{0}/{1}/configuration/stanza/{2}/entry_name/{3}".format(uri, reverseproxy_id, stanza_id, entry_id)

            # Replace % with %25 if it is not encoded already
            import re
            ruri = re.sub("%(?![0-9a-fA-F]{2})", "%25", f_uri)
            # URL encode
            try:
                # Assume Python3 and import package
                from urllib.parse import quote
            except ImportError:
                # Now try to import Python2 package
                from urllib import quote

            full_uri = quote(ruri)

            return isamAppliance.invoke_delete(
                "Deleting all values from a configuration entry - Reverse Proxy", full_uri)

    return isamAppliance.create_return_object()


def update(isamAppliance, reverseproxy_id, stanza_id, entry_id, value_id, check_mode=False, force=False):
    """
   Updating a configuration entry or entries by stanza - Reverse Proxy
   """
    if force is False:
        exists, update_required, cur_value = _check(isamAppliance, reverseproxy_id, stanza_id, entry_id, value_id)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            # URL being encoded primarily to handle request-log-format that has "%" values in them
            f_uri = "{0}/{1}/configuration/stanza/{2}/entry_name/{3}".format(uri, reverseproxy_id,
                                                                             stanza_id, entry_id)
            # Replace % with %25 if it is not encoded already
            import re
            ruri = re.sub("%(?![0-9a-fA-F]{2})", "%25", f_uri)
            # URL encode
            try:
                # Assume Python3 and import package
                from urllib.parse import quote
            except ImportError:
                # Now try to import Python2 package
                from urllib import quote

            full_uri = quote(ruri)
            return isamAppliance.invoke_put(
                "Updating a configuration entry or entries by stanza - Reverse Proxy",
                full_uri,
                {
                    'value': value_id
                })

    return isamAppliance.create_return_object()


def _check(isamAppliance, reverseproxy_id, stanza_id, entry_id, value_id):
    """
    Check if entry/value exists and if exists then if update is required
    """
    try:
        ret_obj = get(isamAppliance, reverseproxy_id, stanza_id, entry_id)
        exists = True
        update_required = False
        value = ret_obj['data'][entry_id]
    except:
        return False, True, None  # Exception means entry / stanza not found

    logger.info("Entry found in rp:{0}, stanza:{1}, entryid:{2}, value:{3}".format(reverseproxy_id,
                                                                                   stanza_id,
                                                                                   entry_id,
                                                                                   value))
    logger.debug("Existing Value(s): {0}".format(value))
    logger.debug("Value to update  : {0}".format(value_id))

    if isinstance(value_id, list):
        if value != value_id:  # Comparing list with no sorting... sequence of values is of importance
            logger.debug("Value arrays do not match!")
            update_required = True
    else:  # assuming base string provided for value_id
        if len(value) == 1:
            if str(value_id) != str(value[0]):
                logger.debug("Single value do not match!")
                update_required = True
                exists = False  # to satisfy delete call
        else:  # base string will not match a zero length array or multiple values in it
            logger.debug("Current non-single value does not match provided single value!")
            update_required = True
            if value_id not in value:
                logger.debug("Current non-single value does not contain provided single value!")
                exists = False

    return exists, update_required, value


def compare(isamAppliance1, isamAppliance2, reverseproxy_id, stanza_id, reverseproxy_id2=None):
    """
   Compare stanza/entries in two appliances reverse proxy configuration
   """
    if reverseproxy_id2 is None or reverseproxy_id2 == '':
        reverseproxy_id2 = reverseproxy_id

    ret_obj1 = get_all(isamAppliance1, reverseproxy_id, stanza_id)
    ret_obj2 = get_all(isamAppliance2, reverseproxy_id2, stanza_id)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging

logger = logging.getLogger(__name__)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving all Junction Mapping
    """
    return isamAppliance.invoke_get("Retrieving all Junction Mapping",
                                    "/wga/jmt_config")


def get(isamAppliance, id, check_mode=False, force=False):
    """
    Retrieve a Junction Mapping
    """
    return isamAppliance.invoke_get("Retrieve a Junction Mapping",
                                    "/wga/jmt_config/{0}".format(id))


def _get_template(isamAppliance):
    """
    Retrieve a Junction Mapping Template
    """
    return isamAppliance.invoke_get("Retrieve a Junction Mapping Template",
                                    "/isam/wga_templates/jmt_template.json")


def add(isamAppliance, id, jmt_config_data=None, check_mode=False, force=False):
    """
    Add a Junction Mapping
    """
    # Add default template if not specified
    if jmt_config_data is None:
        ret_obj = _get_template(isamAppliance)
        jmt_config_data = ret_obj['data']['id']

    if force is True or _check(isamAppliance, id) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Add a Junction Mapping",
                "/wga/jmt_config",
                {
                    "name": id,
                    "jmt_config_data": jmt_config_data
                })

    return isamAppliance.create_return_object()


def delete(isamAppliance, id, check_mode=False, force=False):
    """
    Deleting a Junction Mapping
    """
    if force is True or _check(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a Junction Mapping",
                "/wga/jmt_config/{0}".format(id))

    return isamAppliance.create_return_object()


def rename(isamAppliance, id, new_name, check_mode=False, force=False):
    """
    Rename a Junction Mapping
    """
    if force is True or _check(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Rename a Junction Mapping",
                "/wga/jmt_config/{0}".format(id),
                {
                    'id': id,
                    'new_name': new_name
                })

    return isamAppliance.create_return_object()


def update(isamAppliance, id, jmt_config_data, check_mode=False, force=False):
    """
    Update a Junction Mapping
    """
    warnings = []
    update_required = False
    ret_obj_content = get(isamAppliance, id)
    if ret_obj_content['data'] == {}:
        warnings.append("Junction Mapping {} not found. Skipping update.".format(id))
    # Having to strip whitespace to get a good comparison (suspect carriage returns added after save happens)
    elif (ret_obj_content['data']['contents']).strip() != (jmt_config_data).strip():
        update_required = True

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Update a Junction Mapping",
                "/wga/jmt_config/{0}".format(id),
                {
                    'id': id,
                    'jmt_config_data': jmt_config_data
                }, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def export_file(isamAppliance, id, filename, check_mode=False, force=False):
    """
    Exporting a Junction Mapping
    """
    import os.path

    if force is True or (_check(isamAppliance, id) is True and os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Export a Junction Mapping",
                "/wga/jmt_config/{0}?export".format(id),
                filename)

    return isamAppliance.create_return_object()


def export_template(isamAppliance, filename, check_mode=False, force=False):
    """
    Exporting the JMT configuration file template
    """
    import os.path

    if os.path.exists(filename) is True:
        logger.info("File '{0}' already exists. Skipping export.".format(filename))
        return isamAppliance.create_return_object()

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_get_file(
            "Exporting the JMT configuration file template",
            "/isam/wga_templates/jmt_template?export",
            filename)

    return isamAppliance.create_return_object()


def import_file(isamAppliance, id, filename, check_mode=False, force=False):
    """
    Importing a Junction Mapping
    """
    if force is True or _check(isamAppliance, id) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Importing a Junction Mapping",
                "/wga/jmt_config.js",
                [
                    {
                        'file_formfield': 'jmt_config_file',
                        'filename': filename,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {})

    return isamAppliance.create_return_object()


def _check(isamAppliance, id):
    """
    Check if Junction Mapping already exists
    """
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if obj['id'] == id:
            return True

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Junction Mapping between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['version']
        ret_obj = get(isamAppliance1, obj['id'])
        obj['script'] = ret_obj['data']['contents']
    for obj in ret_obj2['data']:
        del obj['version']
        ret_obj = get(isamAppliance2, obj['id'])
        obj['script'] = ret_obj['data']['contents']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['version'])



import logging
import os.path

from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)
uri = "/wga/query_sitecontents"
requires_modules = ["wga"]
requires_version = None


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving all query site contents file names
    """
    return isamAppliance.invoke_get("Retrieving all query site contents file names",
                                    "{0}".format(uri),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, file_id, check_mode=False, force=False):
    """
    Retrieving the contents of a query site contents file
    """
    return isamAppliance.invoke_get("Retrieving the contents of a query site contents file",
                                    "{0}/{1}".format(uri, file_id),
                                    requires_modules=requires_modules, requires_version=requires_version)


def export_file(isamAppliance, file_id, filename, check_mode=False, force=False):
    """
    Exporting a query site contents file

    """

    if os.path.exists(filename) is True:
        logger.info("File '{0}' already exists.  Skipping export.".format(filename))
        return isamAppliance.create_return_object()

    filefound = False

    ret_obj = get_all(isamAppliance)
    for obj in ret_obj['data']:
        if obj['id'] == file_id:
            filefound = True

    if force is True or filefound is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            return isamAppliance.invoke_get_file(
                "Exporting a query site contents file",
                "{0}/{1}?export".format(uri, file_id), filename,
                requires_modules=requires_modules, requires_version=requires_version
            )
    return isamAppliance.create_return_object()



import logging
import json
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/fido2/relying-parties"
requires_modules = ["mga"]
requires_version = "9.0.7.0"

def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of all FIDO2 Relying Part
    """
    return isamAppliance.invoke_get("Retrieving the list of all FIDO2 Relying Parties", uri,
                                    requires_modules=requires_modules, requires_version=requires_version)

def get(isamAppliance, name, id=None, check_mode=False, force=False):
    """
    Retrieve a specific FIDO2 Relying Party
    """
    if id is None:
        ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
        id = ret_obj['data']

    if id == {}:
        logger.info("FIDO2 Metadata {0} had no match, skipping retrieval.".format(name))
        return isamAppliance.create_return_object()
    else:
        return _get(isamAppliance, id)

def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search FIDO2 Relying Party id by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found FIDO2 relying party {0} id: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj

def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete a FIDO2 Relying Party
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        logger.info("FIDO2 relying party {0} not found, skipping delete.".format(name))
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a FIDO2 relying party",
                "{0}/{1}".format(uri, id),
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()

def set(isamAppliance, name, rpId, fidoServerOptions, relyingPartyOptions, check_mode=False, force=False):
    """
    Create or Update a FIDO2 Relying Party
    """
    if (search(isamAppliance, name=name))['data'] == {}:
        # Force the add - we already know FIDO2 Relying Party does not exist
        logger.info("FIDO2 relying party {0} had no match, requesting to add new one.".format(name))
        return add(isamAppliance, name=name, rpId=rpId, fidoServerOptions=fidoServerOptions, relyingPartyOptions=relyingPartyOptions,
                   check_mode=check_mode, force=True)
    else:
        # Update request
        logger.info("FIDO2 relying party {0} exists, requesting to update.".format(name))
        return update(isamAppliance, name=name, rpId=rpId, fidoServerOptions=fidoServerOptions, relyingPartyOptions=relyingPartyOptions,
                      check_mode=check_mode, force=force)

def add(isamAppliance, name, rpId, fidoServerOptions, relyingPartyOptions, check_mode=False, force=False):
    """
    Create a new FIDO2 Relying Party
    """
    if force is False:
        ret_obj = search(isamAppliance, name)
    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Create a new FIDO2 relying party", uri,
                {
                    "name": name,
                    "rpId": rpId,
                    "fidoServerOptions": fidoServerOptions,
                    "relyingPartyOptions": relyingPartyOptions
                }, requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()

def update(isamAppliance, name, rpId, fidoServerOptions, relyingPartyOptions, check_mode=False, force=False):
    """
    Update a specific FIDO2 Relying Party
    """
    rp_id, update_required, json_data = _check(isamAppliance, name=name, rpId=rpId, fidoServerOptions=fidoServerOptions, relyingPartyOptions=relyingPartyOptions)
    if rp_id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot update data for unknown FIDO2 relying party: {0}".format(name))

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specific FIDO2 relying party",
                "{0}/{1}".format(uri, rp_id), json_data,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()

def _get(isamAppliance, id):
    return isamAppliance.invoke_get("Retrieve a specific FIDO2 relying party by id",
                                    f"{uri}/{id}")

def _extract_filename(upload_filename):
    """
    Extract filename from fully qualified path to use if no filename provided
    """
    import os.path
    return os.path.basename(upload_filename)

def _check(isamAppliance, name, rpId, fidoServerOptions, relyingPartyOptions):
    """
    Check and return True if update needed

    :param isamAppliance:
    :param name:
    :param rpId:
    :param fidoServerOptions:
    :param relyingPartyOptions:
    :return:
    """
    update_required = False
    json_data = {}
    ret_obj = get(isamAppliance, name)
    if ret_obj['data'] == {}:
        logger.info("FIDO2 relying party not found, returning no update required.")
        return None, update_required, json_data
    else:
        rp_id = ret_obj['data']['id']

        if name is not None:
            json_data["name"] = name

        if rpId is not None:
            json_data["rpId"] = rpId

        if fidoServerOptions is not None:
            json_data["fidoServerOptions"] = fidoServerOptions

        if relyingPartyOptions is not None:
            json_data["relyingPartyOptions"] = relyingPartyOptions

        # Remove id as this is server specific dynamic data. Not required for configuration comparison
        del ret_obj['data']['id']

        import ibmsecurity.utilities.tools
        sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
        logger.debug("Sorted input: {0}".format(sorted_json_data))
        sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj['data'])
        logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
        if sorted_ret_obj != sorted_json_data:
            logger.info("Changes detected, update needed.")
            update_required = True
        else:
            logger.info("No Changes detected, update NOT required.")
    return rp_id, update_required, json_data



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)

uri = "/snmpd/v1"


def get(isdsAppliance, check_mode=False, force=False):
    """
    Get SNMP Monitoring v1/2
    """
    return isdsAppliance.invoke_get("Get SNMP Monitoring Setup", uri)


def set_v1v2(isdsAppliance, community, port=161, check_mode=False, force=False):
    """
    Set SNMP Monitoring v1/2
    """
    if force is True or _checkv1_2(isdsAppliance, community, port) is False:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True)
        else:
            return isdsAppliance.invoke_put(
                "Updating SNMP Monitoring", uri,
                {
                    "snmpv1v2c": {"community": community},
                    "enabled": True,
                    "port": port
                })

    return isdsAppliance.create_return_object()


def _checkv1_2(isdsAppliance, community, port):
    ret_obj = get(isdsAppliance)

    if ret_obj['data'][0]['enabled'] is False:
        logger.info("SNMP Monitoring is not enabled")
        return False

    if ret_obj['data'][0]['snmpv1v2c']['community'] != community:
        logger.info("SNMP Monitoring community is different")
        return False

    if ret_obj['data'][0]['port'] != port:
        logger.info("SNMP Monitoring port is different")
        return False

    return True


def set_v3(isdsAppliance, securityLevel, securityUser, authPassword, authProtocol, privacyPassword, privacyProtocol,
           port=161, check_mode=False, force=False):
    """
    Set SNMP Monitoring v3
    """
    if force is True or _checkV3(isdsAppliance, securityLevel, securityUser, authPassword, authProtocol,
                                 privacyPassword, privacyProtocol, port) is False:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True)
        else:
            return isdsAppliance.invoke_put(
                "Updating SNMP Monitoring", uri,
                {
                    "snmpv3": {"securityLevel": securityLevel,
                               "securityUser": securityUser,
                               "authPassword": authPassword,
                               "authProtocol": authProtocol,
                               "privacyPassword": privacyPassword,
                               "privacyProtocol": privacyProtocol
                               },
                    "enabled": True,
                    "port": port
                })

    return isdsAppliance.create_return_object()


def _checkV3(isdsAppliance, securityLevel, securityUser, authPassword, authProtocol, privacyPassword, privacyProtocol,
             port):
    ret_obj = get(isdsAppliance)

    if ret_obj['data'][0]['enabled'] is False:
        logger.info("SNMP Monitoring is not enabled")
        return False

    if ret_obj['data'][0]['snmpv3']['securityLevel'] != securityLevel:
        logger.info("SNMP Monitoring security level is different")
        return False

    if ret_obj['data'][0]['snmpv3']['securityUser'] != securityUser:
        logger.info("SNMP Monitoring security user is different")
        return False

    if ret_obj['data'][0]['snmpv3']['authPassword'] != authPassword:
        logger.info("SNMP Monitoring auth password is different")
        return False

    if ret_obj['data'][0]['snmpv3']['authProtocol'] != authProtocol:
        logger.info("SNMP Monitoring auth protocol is different")
        return False

    if ret_obj['data'][0]['snmpv3']['privacyPassword'] != privacyPassword:
        logger.info("SNMP Monitoring privacy password is different")
        return False

    if ret_obj['data'][0]['snmpv3']['privacyProtocol'] != privacyProtocol:
        logger.info("SNMP Monitoring privacy protocol is different")
        return False

    if ret_obj['data'][0]['port'] != port:
        logger.info("SNMP Monitoring port is different")
        return False

    return True


def disable(isdsAppliance, check_mode=False, force=False):
    """
    Disable SNMP Monitoring
    """
    if force is False:
        ret_obj = get(isdsAppliance)

    if force is True or ret_obj['data'][0]['enabled'] is True:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True)
        else:
            return isdsAppliance.invoke_put(
                "Disabling SNMP Monitoring", uri,
                {
                    "enabled": False
                })

    return isdsAppliance.create_return_object()


def compare(isdsAppliance1, isdsAppliance2):
    ret_obj1 = get(isdsAppliance1)
    ret_obj2 = get(isdsAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
import ibmsecurity.utilities.tools
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/isam/rsyslog_forwarder"
requires_modules = None
requires_version = "9.0.2.1"

try:
    basestring
except NameError:
    basestring = (str, bytes)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the current remote syslog forwarding policy
    """
    return isamAppliance.invoke_get("Retrieve the current remote syslog forwarding policy", uri,
                                    requires_modules=requires_modules, requires_version=requires_version)


def _find_forwarder(ret_obj, server, port, protocol):
    '''
    Will return None or the forwarder object found
    index returned will be length of array if no match
    '''
    existing_forwarder = None
    i = 0
    for obj in ret_obj['data']:
        if obj['server'] == server and obj['port'] == port and obj['protocol'] == protocol:
            existing_forwarder = obj
            logger.debug("Found Forwarder: {0}".format(obj))
            break
        i += 1

    return existing_forwarder, i


def get(isamAppliance, server, port, protocol, check_mode=False, force=False):
    """
    Retrieve a specific remote syslog forwarder
    """
    ret_obj = get_all(isamAppliance, check_mode, force)

    if isinstance(port, basestring):
        port = int(port)

    return_obj = isamAppliance.create_return_object()
    return_obj['data'], i = _find_forwarder(ret_obj, server, port, protocol)
    warnings = []
    if return_obj['data'] == None:
        warnings.append("No entry found for server {} port {} and protocol {}.".format(server, port, protocol))
        return_obj['warnings'] = warnings

    return return_obj


def delete(isamAppliance, server, port, protocol, check_mode=False, force=False):
    """
    Remove a specific remote syslog forwarder
    """
    ret_obj = get_all(isamAppliance, check_mode, force)

    if isinstance(port, basestring):
        port = int(port)

    existing_forwarder, i = _find_forwarder(ret_obj, server, port, protocol)

    json_to_post = ret_obj['data']
    if existing_forwarder is not None:
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            del json_to_post[i]
            return _update_forwarder_policy(isamAppliance, json_to_post)

    return isamAppliance.create_return_object()


def set(isamAppliance, server, port, protocol='udp', debug=False, keyfile=None, ca_certificate=None,
        client_certificate=None, permitted_peers=None, sources=[], format=None, check_mode=False, force=False):
    ret_obj = get_all(isamAppliance, check_mode, force)

    if isinstance(port, basestring):
        port = int(port)

    warnings = []
    existing_forwarder, i = _find_forwarder(ret_obj, server, port, protocol)
    if existing_forwarder is not None and sources == [] and existing_forwarder['sources'] != sources:
        sources = existing_forwarder['sources']
        warnings.append("No sources provided, using existing sources to set forwarder.")

    json_data = {
        'server': server,
        'port': port,
        'protocol': protocol,
        'debug': debug,
        'sources': sources
    }
    if keyfile is not None and keyfile != '':
        json_data['keyfile'] = keyfile
    if ca_certificate is not None and ca_certificate != '':
        json_data['ca_certificate'] = ca_certificate
    if client_certificate is not None and client_certificate != '':
        json_data['client_certificate'] = client_certificate
    if permitted_peers is not None and permitted_peers != '':
        json_data['permitted_peers'] = permitted_peers
    update_required = False
    # Check of the given server/port/protocol exist - if not then we add it
    if existing_forwarder is None or force is True:
        if format is not None:
            if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "10.0.2.0") < 0:
                warnings.append(
                    "Appliance at version: {0}, format is not supported. Needs 10.0.2.0 or higher. Ignoring format for this call.".format(
                        isamAppliance.facts["version"]))
            else:
                json_data["format"] = format
        json_to_post = ret_obj['data']
        json_to_post.append(json_data)
        update_required = True
        warnings.append("existing_forwarder is None")
    else:
        if format is not None:
            if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "10.0.2.0") < 0:
                warnings.append(
                    "Appliance at version: {0}, format is not supported. Needs 10.0.2.0 or higher. Ignoring format for this call.".format(
                        isamAppliance.facts["version"]))
            else:
                json_data["format"] = format
        elif 'format' in ret_obj['data'][i]:
            del ret_obj['data'][i]['format']
        sorted_json_data = tools.json_sort(json_data)
        logger.debug("Sorted input: {0}".format(sorted_json_data))
        sorted_ret_obj = tools.json_sort(existing_forwarder)
        logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
        if sorted_ret_obj != sorted_json_data:
            logger.info("Changes detected, update needed.")
            ret_obj['data'][i] = json_data
            json_to_post = ret_obj['data']
            update_required = True

    if update_required:
        if check_mode:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return _update_forwarder_policy(isamAppliance, json_to_post, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def _update_forwarder_policy(isamAppliance, json_to_post, warnings=[]):
    return isamAppliance.invoke_put(
        "Update the current remote syslog forwarding policy", uri,
        json_to_post, requires_modules=requires_modules,
        requires_version=requires_version, warnings=warnings)


def compare(isamAppliance1, isamAppliance2):
    """
    Compare forwarder Policies between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/mga/runtime_profile/local/v1"
requires_modules = ["mga", "federation"]
requires_version = "8.0.0.0"
requires_model = "Appliance"

def get(isamAppliance, check_mode=False, force=False):
    """
    Get runtime status
    """
    return isamAppliance.invoke_get("Retrieving runtime status",
                                    "/mga/runtime_profile/v1",
                                    requires_modules=requires_modules, requires_version=requires_version,requires_model=requires_model)


def execute(isamAppliance, operation="restart", check_mode=False, force=False):
    """
    Execute an operation (start, stop or restart) on runtime
    """
    check_value, warnings = _check(isamAppliance,operation)

    # Reload function is new to ISAM v9.0.2.0
    if operation == "reload" and tools.version_compare(isamAppliance.facts["version"], "9.0.2.0") < 0:
        warnings.append(
            "Appliance is at version {0}, reload requires atleast v9.0.2.0".format(isamAppliance.facts['version']))
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or check_value is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Executing an operation on runtime", uri,
                {
                    'operation': operation
                }, requires_modules=requires_modules, requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, operation):
    """
    Check if operation needs to be applied to process
    """
    ret_obj = get(isamAppliance)
    warnings = ret_obj['warnings']

    if 'restart_required' in ret_obj['data'] and ret_obj['data']['restart_required'] is True and operation == "restart":
        logger.info("Liberty Runtime process requires a restart")
        return True,warnings
    elif 'return_code' in ret_obj['data'] and ret_obj['data']['return_code'] == 0 and operation == "stop":
        logger.info("Liberty Runtime process is running and can be stopped")
        return True,warnings
    elif 'return_code' in ret_obj['data'] and ret_obj['data']['return_code'] == 1 and operation == "start":
        logger.info("Liberty Runtime process is not running and can be started")
        return True,warnings
    elif 'reload_required' in ret_obj['data'] and ret_obj['data']['reload_required'] is True and operation == "reload":
        logger.info("Liberty Runtime process requires reload")
        return True,warnings
    else:
        logger.info(
            "Executing {0} on Liberty Runtime process not required or will not work. Use force if needed.".format(
                operation))
        return False,warnings



import logging
import time
from ibmsecurity.isvg import notifications

logger = logging.getLogger(__name__)


def restart(isvgAppliance, check_mode=False, force=False):
    """
    Restart LMI
    """
    operation = "restart"
    check_value, warnings = _check(isvgAppliance, operation)

    if force is True or check_value is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isvgAppliance.invoke_post(
                "Executing an operation on LMI", "/restarts/restart_server",
                {
                    'operation': operation
                }, warnings=warnings)


def _check(isvgAppliance, operation):
    """
    Check if operation needs to be applied to LMI
    """
    ret_obj = notifications.get(isvgAppliance)
    warnings = ret_obj['warnings']

    if 'items' in ret_obj['data']:
        for item in ret_obj['data']['items']:
            if 'message' in item and item['message'] == 'lmi_restart_reqd' and operation == "restart":
                logger.info("Identity Manager server requires a restart")
                return True,warnings
    else:
        logger.info(
            "Executing {0} on LMI not required or will not work. Use force if needed.".format(
                operation))
        return False,warnings

    return False,warnings


import logging
import ibmsecurity.utilities.tools as tools
import json

try:
    basestring
except NameError:
    basestring = (str, bytes)

logger = logging.getLogger(__name__)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the list of federated directories
    """
    return isamAppliance.invoke_get("Retrieving the list of federated directories",
                                    "/isam/runtime_components/federated_directories/v1")


def get(isamAppliance, id, check_mode=False, force=False):
    """
    Retrieving the details for a particular federated directory
    """
    return isamAppliance.invoke_get("Retrieving the details for a particular federated directory",
                                    "/isam/runtime_components/federated_directories/{0}/v1".format(id))


def set(isamAppliance, id, hostname, port, bind_dn, bind_pwd, suffix, use_ssl=False, client_cert_label=None,
        ignore_if_down=False,
        check_mode=False, force=False):
    if _exists(isamAppliance, id) is False:
        return add(isamAppliance, id=id, hostname=hostname, port=port, bind_dn=bind_dn, bind_pwd=bind_pwd,
                   suffix=suffix, use_ssl=use_ssl, client_cert_label=client_cert_label,
                   ignore_if_down=ignore_if_down,
                   check_mode=check_mode,
                   force=True)
    else:
        return update(isamAppliance, id=id, hostname=hostname, port=port, bind_dn=bind_dn, bind_pwd=bind_pwd,
                      suffix=suffix, use_ssl=use_ssl, client_cert_label=client_cert_label,
                      ignore_if_down=ignore_if_down,
                      check_mode=check_mode,
                      force=force)


def add(isamAppliance, id, hostname, port, bind_dn, bind_pwd, suffix, use_ssl=False, client_cert_label=None,
        ignore_if_down=False,
        check_mode=False, force=False):
    """
    Create a new federated directory
    """
    if (isinstance(suffix, basestring)):
        import ast
        suffix = ast.literal_eval(suffix)

    if force is True or _exists(isamAppliance, id) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                'id': id,
                'hostname': hostname,
                'port': port,
                'bind_dn': bind_dn,
                'bind_pwd': bind_pwd,
                'use_ssl': use_ssl,
                'suffix': suffix
            }

            if ignore_if_down and tools.version_compare(isamAppliance.facts["version"], "10.0.4") >= 0:
                json_data['ignore_if_down'] = ignore_if_down

            # Do not pass if there is no value - call fails otherwise
            if client_cert_label is not None:
                json_data['client_cert_label'] = client_cert_label
            return isamAppliance.invoke_post(
                "Create a new federated directory",
                "/isam/runtime_components/federated_directories/v1", json_data)

    return isamAppliance.create_return_object()


def update(isamAppliance, id, hostname, port, bind_dn, bind_pwd, suffix, use_ssl=False, client_cert_label=None,
           ignore_if_down=False,
           check_mode=False, force=False):
    """
    Update an existing federated directory
    """
    if force is True or (
            _exists(isamAppliance, id) and _check(isamAppliance, id, hostname, port, bind_dn, bind_pwd,
                                                          use_ssl, client_cert_label, suffix, ignore_if_down) is False):

        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                'hostname': hostname,
                'port': port,
                'bind_dn': bind_dn,
                'bind_pwd': bind_pwd,
                'use_ssl': use_ssl,
                'suffix': suffix
            }
            if ignore_if_down and tools.version_compare(isamAppliance.facts["version"], "10.0.4") >= 0:
                json_data['ignore_if_down'] = ignore_if_down
            # Do not pass if there is no value - call fails otherwise
            if client_cert_label is not None:
                json_data['client_cert_label'] = client_cert_label
            return isamAppliance.invoke_put(
                "Update an existing federated directory",
                "/isam/runtime_components/federated_directories/{0}/v1".format(id), json_data)

    return isamAppliance.create_return_object()


def delete(isamAppliance, id, check_mode=False, force=False):
    """
    Remove an existing federated directory
    """
    if force is True or _exists(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Remove an existing federated directory",
                "/isam/runtime_components/federated_directories/{0}/v1".format(id))

    return isamAppliance.create_return_object()


def _exists(isamAppliance, id):
    """
    Check if federated directory exists

    :param isamAppliance:
    :param id:
    :return:
    """
    exists = False
    ret_obj = get_all(isamAppliance)

    for snmp in ret_obj['data']:
        if snmp['id'] == id:
            exists = True
            break

    return exists


def _check(isamAppliance, id, hostname, port, bind_dn, bind_pwd, use_ssl, client_cert_label, suffix, ignore_if_down=False):
    """
    Check if parameters match given federated directory

    Note: This does not check bind_pwd

    Returns True if it exists and is the same
    """
    if _exists(isamAppliance, id):
        ret_obj = get(isamAppliance, id)
    else:
        return False

    set_value = {
        'id': id,
        'hostname': hostname,
        'port': str(port),
        'bind_dn': bind_dn,
        'use_ssl': use_ssl,
        'suffix': suffix
    }
    if use_ssl is True:
        set_value['client_cert_label'] = client_cert_label
    if ignore_if_down and tools.version_compare(isamAppliance.facts["version"], "10.0.4") >= 0:
        set_value['ignore_if_down'] = ignore_if_down

    newEntriesJSON = json.dumps(set_value, skipkeys=True, sort_keys=True)
    logger.debug("\nSorted New Federated Directory {0}: {1}".format(id, newEntriesJSON))
    currentEntriesJSON = json.dumps(ret_obj['data'], skipkeys=True, sort_keys=True)
    logger.debug("\nSorted Existing Federated Directory {0}: {1}".format(id, currentEntriesJSON))

    if newEntriesJSON == currentEntriesJSON:
        return True
    else:
        return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare federated directory stanze entries between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging

logger = logging.getLogger(__name__)

uri = "/iam/access/v8/otp/config/rsa"
requires_modules = ["mga"]
requires_version = "8.0.0.0"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of configuration files for RSA
    """
    return isamAppliance.invoke_get("Retrieve a list of configuration files for RSA",
                                    "{0}".format(uri),
                                    requires_modules=requires_modules, requires_version=requires_version)



import logging
import zipfile
import shutil

from ibmsecurity.utilities.tools import get_random_temp_dir

logger = logging.getLogger(__name__)
requires_model = "Appliance"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Get list of all certificate databases
    """
    return isamAppliance.invoke_get("Retrieving all current certificate database names",
                                    "/isam/ssl_certificates")


def get(isamAppliance, cert_dbase_id, check_mode=False, force=False):
    """
    Retrieving the SSL certificate database details
    """
    return isamAppliance.invoke_get("Retrieving the SSL certificate database details",
                                    "/isam/ssl_certificates/{0}/details".format(cert_dbase_id),
                                    requires_model=requires_model)


def create(isamAppliance, kdb_name, type='kdb', token_label=None, passcode=None, hsm_type=None, ip=None, port=None,
           kneti_hash=None, esn=None, rfs=None, rfs_port=None, rfs_auth=None, safenet_pw=None, check_mode=False,
           force=False):
    """
    Create a certificate database
    """
    if force is True or _check(isamAppliance, kdb_name) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Creating certificate database '{0}'".format(kdb_name),
                "/isam/ssl_certificates",
                {
                    "kdb_name": kdb_name,
                    "type": type,
                    "token_label": token_label,
                    "passcode": passcode,
                    "hsm_type": hsm_type,
                    "ip": ip,
                    "port": port,
                    "kneti_hash": kneti_hash,
                    "esn": esn,
                    "rfs": rfs,
                    "rfs_port": rfs_port,
                    "rfs_auth": rfs_auth,
                    "safenet_pw": safenet_pw
                })

    return isamAppliance.create_return_object()


def delete(isamAppliance, cert_dbase_id, check_mode=False, force=False):
    """
    Delete a certificate database
    """
    if force is True or _check(isamAppliance, cert_dbase_id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a certificate database",
                "/isam/ssl_certificates/{0}".format(cert_dbase_id))

    return isamAppliance.create_return_object()


def export_db(isamAppliance, cert_id, filename, check_mode=False, force=False):
    """
    Export a certificate database
    """
    import os.path

    if force is True or (_check(isamAppliance, cert_id) is True and os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Export a certificate database",
                "/isam/ssl_certificates/{0}?export".format(cert_id),
                filename)

    return isamAppliance.create_return_object()


def import_db(isamAppliance, kdb, stash, zip=None, check_mode=False, force=False):
    """
    Import certificate database
    """
    # Grab the filename to use as identifier (strip path and extension)
    import os.path
    
    tmpdir=None
    if zip != None:
        with zipfile.ZipFile(zip,"r") as zip_ref:
            tmpdir = get_random_temp_dir()
            zip_ref.extractall(tmpdir)
            kdb = tmpdir + '/' + kdb
            stash = tmpdir + '/' + stash
    
    kdb_id = os.path.basename(kdb)
    kdb_id = os.path.splitext(kdb_id)[0]

    try:
        if force is True or _check(isamAppliance, kdb_id) is False:
            if check_mode is True:            
                return isamAppliance.create_return_object(changed=True)
            else:
                return isamAppliance.invoke_post_files(
                    "Import certificate database",
                    "/isam/ssl_certificates",
                    [
                        {
                            'file_formfield': 'kdb',
                            'filename': kdb,
                            'mimetype': 'application/octet-stream'
                        },
                        {
                            'file_formfield': 'stash',
                            'filename': stash,
                            'mimetype': 'application/octet-stream'
                        }
                    ],
                    {})
    finally:
        if tmpdir != None:
            shutil.rmtree(tmpdir)
    return isamAppliance.create_return_object()


def rename(isamAppliance, cert_id, new_name, check_mode=False, force=False):
    """
    Rename a certificate database
    """
    if force is True or _check(isamAppliance, cert_id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Renaming a certificate database",
                "/isam/ssl_certificates/{0}".format(cert_id),
                {
                    "new_name": new_name
                })

    return isamAppliance.create_return_object()


def set(isamAppliance, cert_id, description, check_mode=False, force=False):
    """
    Set description for a certificate database
    """
    desc_match = True  # This will remain True even when cert db is not found!
    if force is False:
        ret_obj = get_all(isamAppliance)
        for certdb in ret_obj['data']:
            if certdb['id'] == cert_id:
                if certdb['description'] != description:
                    desc_match = False
                break

    if force is True or desc_match is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Set description for a certificate database",
                "/isam/ssl_certificates/{0}".format(cert_id),
                {
                    "description": description
                })

    return isamAppliance.create_return_object()


def _check(isamAppliance, id):
    """
    Check if certificate database already exists
    """
    ret_obj = get_all(isamAppliance)

    for certdb in ret_obj['data']:
        if certdb['id'] == id:
            return True

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare certificate databases between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    # Version is last modified time of file - not worth comparing
    for cert_db in ret_obj1['data']:
        del cert_db['version']
    for cert_db in ret_obj2['data']:
        del cert_db['version']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['version'])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)
artifact_type = "Attributes"

# URI for this module
uri = "/iam/access/v8/attributes"


def get_all(isamAppliance, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of Attributes
    """
    return isamAppliance.invoke_get("Retrieve a list of attributes",
                                    "{0}/{1}".format(uri, tools.create_query_string(filter=filter, sortBy=sortBy)))


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific Attribute
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    artifact_id = ret_obj['data']

    if artifact_id == {}:
        logger.info("Attribute {0} had no match, skipping retrieval.".format(name))
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get("Retrieve a specific attribute",
                                        "{0}/{1}".format(uri, artifact_id))


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search attribute id by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        logger.info("Obj is: {0}".format(obj))
        if obj['name'] == name:
            logger.info("Found attribute {0} id: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def set(isamAppliance, name, description, attributeURI, type, category, datatype, predefined, issuer, matcher,
        storageDomain,
        new_name=False, check_mode=False, force=False):
    """
    Creating or Modifying a Attribute
    """
    ret_obj = search(isamAppliance, name)

    if ret_obj['data'] == {}:
        logger.info('Adding "{1}" as a new {0}'.format(artifact_type, name))
        return add(isamAppliance, name, description, attributeURI, type, category, datatype, predefined, issuer,
                   matcher, storageDomain,
                   check_mode, True)
    else:
        logger.info('Update for {0} "{1}"'.format(artifact_type, name))
        return update(isamAppliance, name, description, attributeURI, type, category, datatype, predefined, issuer,
                      matcher, storageDomain,
                      new_name, check_mode, force)


def add(isamAppliance, name, description, attributeURI, type, category, datatype, predefined, issuer, matcher,
        storageDomain,
        check_mode=False, force=False):
    """
    Create a new Attribute
    """
    if force is False:
        ret_obj = search(isamAppliance, name)

    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                "name": name,
                "description": description,
                "uri": attributeURI,
                "type": type,
                "category": category,
                "datatype": datatype,
                "predefined": predefined,
                "issuer": issuer,
                "matcher": matcher,
                "storageDomain": storageDomain
            }
            return isamAppliance.invoke_post("Create a new attribute", uri, json_data)

    return isamAppliance.create_return_object()


def update(isamAppliance, name, description, attributeURI, type, category, datatype, predefined, issuer, matcher,
           storageDomain,
           new_name=False, check_mode=False, force=False):
    """
    Update a specified Attribute

    """
    artifact_id, update_required, json_data = _check(isamAppliance, name, description, attributeURI,
                                                     type, category, datatype, predefined, issuer, matcher,
                                                     storageDomain, new_name)
    if artifact_id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot update data for unknown attribute: {0}".format(name))

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified attribute",
                "{0}/{1}".format(uri, artifact_id), json_data)

    return isamAppliance.create_return_object()


def _check(isamAppliance, name, description, attributeURI, type, category, datatype, predefined, issuer, matcher,
           storageDomain,
           new_name=False):
    """
    Check and return True if update needed
    Added logic to check for 'predefined' objects, which cannot be updated
    """
    update_required = False
    json_data = {
        "name": name,
        "description": description,
        "uri": attributeURI,
        "type": type,
        "category": category,
        "datatype": datatype,
        "predefined": predefined,
        "issuer": issuer,
        "matcher": matcher,
        "storageDomain": storageDomain
    }
    logger.warning(" New Name {0}.".format(new_name))

    ret_obj = get(isamAppliance, name);
    if ret_obj['data'] == {}:
        logger.warning(" not found, returning no update required.")
        return None, update_required, json_data
    elif ret_obj['data']['predefined']:
        logger.warning("A predefined {0} cannot be updated, returning no update required.".format(artifact_type))
        return None, update_required, json_data
    else:
        artifact_id = ret_obj['data']['id']
        if new_name != False:
            logger.warning(" New Name 1 {0}.".format(new_name))
            json_data['name'] = new_name

        del ret_obj['data']['id']
        del ret_obj['data']['tags']
        import ibmsecurity.utilities.tools
        sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
        logger.debug("Sorted input: {0}".format(sorted_json_data))
        sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj['data'])
        logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
        if sorted_ret_obj != sorted_json_data:
            logger.info("Changes detected, update needed.")
            update_required = True

    return artifact_id, update_required, json_data


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete an Obligation
    """
    ret_obj = get(isamAppliance, name)

    if ret_obj['data'] == {}:
        logger.info('{0} "{1}" not found, skipping delete.'.format(artifact_type, name))
    elif ret_obj['data']["predefined"]:
        logger.info('{0} "{1}" is predefined, skipping delete.'.format(artifact_type, name))
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            logger.info('Deleting {0} "{1}"'.format(artifact_type, name))
            return isamAppliance.invoke_delete(
                "Delete a Attribute",
                "{0}/{1}".format(uri, ret_obj['data']['id']))

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Policy Sets between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
    for obj in ret_obj2['data']:
        del obj['id']

    return tools.json_compare(ret_obj1, ret_obj2,
                              deleted_keys=['id'])



import logging
import ibmsecurity.utilities.tools
from ibmsecurity.utilities import tools
import ibmsecurity.isam.web.reverse_proxy.junctions

logger = logging.getLogger(__name__)

uri = "/wga/reverseproxy"


def search(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port):
    ret_obj_new = isamAppliance.create_return_object()
    ret_obj = ibmsecurity.isam.web.reverse_proxy.junctions.get(isamAppliance, reverseproxy_id, junction_point)
    for s in ret_obj['data']['servers']:
        logger.debug("Servers in Junction server: {0} port: {1}".format(s['server_hostname'], s['server_port']))
        if str(server_hostname) == str(s['server_hostname']) and str(server_port) == str(s['server_port']):
            ret_obj_new['data'] = s['server_uuid']
            break

    return ret_obj_new


def add(isamAppliance, reverseproxy_id, junction_point, server_hostname, junction_type, server_port, server_dn=None,
        stateful_junction='no', case_sensitive_url='no', windows_style_url='no', virtual_hostname=None,
        virtual_https_hostname=None, query_contents=None, https_port=None, http_port=None, proxy_hostname=None,
        proxy_port=None, sms_environment=None, vhost_label=None, server_uuid=None, priority=None, server_cn=None, check_mode=False, force=False):
    """
    Adding a back-end server to an existing standard or virtual junctions

    :param isamAppliance:
    :param reverseproxy_id:
    :param junctionname:
    :param server_hostname:
    :param junction_type:
    :param server_port:
    :param virtual_hostname:
    :param virtual_https_hostname:
    :param server_dn:
    :param query_contents:
    :param stateful_junction:
    :param case_sensitive_url:
    :param windows_style_url:
    :param https_port:
    :param http_port:
    :param proxy_hostname:
    :param proxy_port:
    :param sms_environment:
    :param vhost_label:
    :param server_uuid:
    :param priority:
    :param server_cn:
    :param check_mode:
    :param force:
    :return:
    """
    # Search for the UUID of the junctioned server
    if force is False:
        ret_obj = search(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port)

    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            jct_srv_json = {
                "junction_point": junction_point,
                "junction_type": junction_type,
                "server_hostname": server_hostname,
                "server_port": server_port,
                "stateful_junction": stateful_junction,
                "case_sensitive_url": case_sensitive_url,
                "windows_style_url": windows_style_url,
            }
            if https_port is not None:
                jct_srv_json["https_port"] = https_port
            if http_port is not None:
                jct_srv_json["http_port"] = http_port
            if proxy_hostname is not None:
                jct_srv_json["proxy_hostname"] = proxy_hostname
            if proxy_port is not None:
                jct_srv_json["proxy_port"] = proxy_port
            if sms_environment is not None:
                jct_srv_json["sms_environment"] = sms_environment
            if vhost_label is not None:
                jct_srv_json["vhost_label"] = vhost_label
            if server_dn is not None:
                jct_srv_json["server_dn"] = server_dn
            if virtual_hostname:
                jct_srv_json["virtual_hostname"] = virtual_hostname
            if virtual_https_hostname is not None:
                jct_srv_json["virtual_https_hostname"] = virtual_https_hostname
            if query_contents is not None:
                jct_srv_json["query_contents"] = query_contents
            if server_uuid is not None and server_uuid != '':
                jct_srv_json["server_uuid"] = server_uuid
            if server_cn is not None:
                if tools.version_compare(isamAppliance.facts["version"], "10.0.2.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, server_cn: {1} is not supported. Needs 10.0.2.0 or higher. Ignoring server_cn for this call.".format(
                            isamAppliance.facts["version"], server_cn))
                    server_cn = None
                else:
                    jct_srv_json["server_cn"] = server_cn
            if priority is not None:
                if tools.version_compare(isamAppliance.facts["version"], "10.0.2.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, priority: {1} is not supported. Needs 10.0.2.0 or higher. Ignoring priority for this call.".format(
                            isamAppliance.facts["version"], priority))
                    priority = None
                else:
                    jct_srv_json["priority"] = priority

            return isamAppliance.invoke_put(
                "Adding a back-end server to an existing standard or virtual junctions",
                "{0}/{1}/junctions".format(uri, reverseproxy_id), jct_srv_json)

    return isamAppliance.create_return_object()


def delete(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port, check_mode=False, force=False):
    """
    Deleting a standard or virtual junction's server

    :param isamAppliance:
    :param reverseproxy_id:
    :param junction_point:
    :param check_mode:
    :param force:
    :return:
    """
    # Search for the UUID of the junctioned server
    if force is False:
        ret_obj = search(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port)

    if force is True or ret_obj['data'] != {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a standard or virtual junction's server",
                "{0}/{1}/junctions?junctions_id={2}&servers_id={3}".format(uri, reverseproxy_id, junction_point,
                                                                           ret_obj['data']))

    return isamAppliance.create_return_object()



import logging
from ibmsecurity.isam.web.reverse_proxy import junctions

logger = logging.getLogger(__name__)
requires_modules = ["wga"]
requires_version = "9.0.6.0"


def config(isamAppliance, instance_id, hostname='127.0.0.1', port=443, username='easuser', password='passw0rd',
           junction="/mga", reuse_certs=False, reuse_acls=False, check_mode=False, force=False):
    """
    Authentication and Context based access configuration for a reverse proxy instance

    :param isamAppliance:
    :param instance_id:
    :param junction:
    :param hostname:
    :param port:
    :param username:
    :param password:
    :param reuse_certs:
    :param reuse_acls:
    :param check_mode:
    :param force:
    :return:
    """
    warnings = [
        "Idempotency logic will check for existence of {} junction. Use force=True to override.".format(junction)]
    if force is True or _check_config(isamAppliance, instance_id, junction) is False:
        json_data = {
            "junction": junction,
            "hostname": hostname,
            "port": port,
            "username": username,
            "password": password,
            "reuse_certs": reuse_certs,
            "reuse_acls": reuse_acls
        }
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post(
                " Authentication and Context based access configuration for a reverse proxy instance",
                "/wga/reverseproxy/{}/authsvc_config".format(instance_id), json_data, warnings=warnings,
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def _check_config(isamAppliance, instance_id, junction):
    """
    Check if the junction for aac already created

    :param isamAppliance:
    :param instance_id:
    :param junction:
    :return:
    """
    ret_obj = junctions.get_all(isamAppliance, instance_id)

    for j in ret_obj['data']:
        if j['id'] == junction:
            logger.info("Junction {} was found - hence aac config must have already executed.".format(junction))
            return True

    return False



import logging
from ibmsecurity.isam.web.reverse_proxy.configuration import entry

logger = logging.getLogger(__name__)
requires_modules = ["wga"]
requires_version = "9.0.3.0"


def config(isamAppliance, instance_id, lmi, runtime, reuse_acls=None, reuse_certs=None, channel=None,
           check_mode=False, force=False):
    """
    MMFA configuration for a reverse proxy instance
    """
    if force is True or _check_config(isamAppliance, instance_id) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                "lmi": lmi,
                "runtime": runtime
            }
            # Add optional values to the JSON
            if reuse_acls is not None and reuse_acls != '':
                json_data['reuse_acls'] = reuse_acls
            if reuse_certs is not None and reuse_certs != '':
                json_data['reuse_certs'] = reuse_certs
            if channel is not None and channel != '':
                json_data['channel'] = channel
            return isamAppliance.invoke_post(
                "MMFA configuration for a reverse proxy instance",
                "/wga/reverseproxy/{0}/mmfa_config".format(instance_id), json_data,
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def unconfig(isamAppliance, instance_id, check_mode=False, force=False):
    """
    MMFA unconfiguration for a reverse proxy instance
    """
    if force is True or _check_config(isamAppliance, instance_id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "MMFA unconfiguration for a reverse proxy instance",
                "/wga/reverseproxy/{0}/mmfa_config".format(instance_id),
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check_config(isamAppliance, instance_id):
    """
    Check webseal config stanza / entry for autoconfig flag to determine if config already done

    :param isamAppliance:
    :param instance_id:
    :return:
    """
    try:
        ret_obj = entry.get(isamAppliance, instance_id, "mmfa-config-info", "autoconfig")
        if ret_obj['data']['autoconfig'] == ["mmfa"]:
            logger.info(
                "WebSEAL entry mmfa-config-info:autoconfig has value {}. 'mmfa' means config already done.".format(
                    ret_obj['data']))
            return True
    except:
        pass

    return False



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)
module_uri = "/iam/access/v8/mmfa-transactions/"
requires_modules = ["mga"]
requires_version = None


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Get information on all existing Transactions
    """
    return isamAppliance.invoke_get("Retrieving a list of transactions", "{0}".format(module_uri),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, id, check_mode=False, force=False):
    """
    Get information on single transaction
    """
    return isamAppliance.invoke_get("Retrieving a transaction", "{0}{1}".format(module_uri, id),
                                    requires_version=requires_version,
                                    requires_modules=requires_modules)


def delete(isamAppliance, id, check_mode=False, force=False):
    """
    Deletes single transaction
    """
    if force is True or _check(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete("Deleting transaction",
                                               "{0}{1}".format(module_uri, id), requires_modules=requires_modules,
                                               requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check(isamAppliance, id=None):
    """
    checks with id exists within isam appliance
    :param isamAppliance:
    :param id:
    :return:
    """
    new_obj = get_all(isamAppliance)

    if id != None:
        for users in new_obj['data']:
            if users['id'] == id:
                return True

    return False



import logging

logger = logging.getLogger(__name__)


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving a list of all server connections
    """
    return isamAppliance.invoke_get("Retrieving a list of all server connections",
                                    "/mga/server_connections/v1")


def test(isamAppliance, connectionType, hostName, port, userName, password, isSSL, keystorename=None, authKeyLabel=None,
         check_mode=False, force=False):
    """
    Test Connection
    """
    if check_mode is True:
        return isamAppliance.create_return_object()

    json_data = {
        'connectionType': connectionType,
        'hostName': hostName,
        'port': port,
        'userName': userName,
        'password': password,
        'isSSL': isSSL,
    }
    if keystorename is not None:
        json_data['keystorename'] = keystorename
    if authKeyLabel is not None:
        json_data['authKeyLabel'] = authKeyLabel

    ret_obj = isamAppliance.invoke_post("Test Connection", "/iam/access/v8/testconnection", json_data)

    # Assumed POST is for changes, this is just a test so indicate no change
    ret_obj['changed'] = False

    return ret_obj


def compare(isamAppliance1, isamAppliance2):
    """
    Compare all Connections between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    # Some connection types may not exist
    if ret_obj1.get('smtp') is not None:
        for obj in ret_obj1['data']['smtp']:
            del obj['uuid']
    if ret_obj1.get('jdbc') is not None:
        for obj in ret_obj1['data']['jdbc']:
            del obj['uuid']
    if ret_obj1.get('ldap') is not None:
        for obj in ret_obj1['data']['ldap']:
            del obj['uuid']
    if ret_obj2.get('smtp') is not None:
        for obj in ret_obj2['data']['smtp']:
            del obj['uuid']
    if ret_obj2.get('jdbc') is not None:
        for obj in ret_obj2['data']['jdbc']:
            del obj['uuid']
    if ret_obj2.get('ldap') is not None:
        for obj in ret_obj2['data']['ldap']:
            del obj['uuid']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['uuid'])
